#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul  6 02:29:17 2022

@author: qy
"""

import numpy as np
import seaborn as sns    
import scipy 
#import matplotlib
import copy
import pandas as pd
import os
import itertools
from matplotlib import pyplot as plt
import warnings

#from auxi import rm_outliers
def rm_outliers(x):
    iqr = scipy.stats.iqr(x)
    outlier_lb = np.quantile(x,0.25)-1.5*iqr
    outlier_ub = np.quantile(x,0.75)+1.5*iqr
    x_shrinkage = x[x > outlier_lb]
    x_shrinkage = x_shrinkage[x_shrinkage<outlier_ub]
    return x_shrinkage#,(outlier_lb,outlier_ub)


def dmn_convergence(group,output,return_fig=None):
    '''
    Visualize the convergence of DMN optimation process.

    Parameters
    ----------
    group : str
        The droplet category to display the optimization process.
    output : path
        The specified output path when calling fucntion 'tomas.fit.dmn'.

    Returns
    -------
    None.

    '''
    log = pd.read_csv(os.path.join(output,group+'.dmnlog.txt'))

    fig, axs =plt.subplots(1,3,figsize=(12,3),dpi=64)

    plt.subplot(1,3,1)
    plt.scatter(log.index, log['loglik'], s=1)
    plt.xlabel('iteration')
    plt.ylabel('logLikelihood')
    plt.title('logLikelihood',fontsize=14)

    plt.subplot(1,3,2)
    plt.scatter(log.index, log['alpha_sum'], s=1)
    plt.xlabel('iteration')
    plt.ylabel('l1-norm of alpha')
    plt.title('precision',fontsize=14)

    plt.subplot(1,3,3)
    plt.scatter(log.index, log['alpha_l2norm_delta'], s=1)
    plt.xlabel('iteration')
    plt.ylabel('delta ||alpha||2')
    plt.title('delta of l2norm of alpha',fontsize=14)

    fig.tight_layout()
    if return_fig is True:
        return fig
    plt.show()




'''
def alpha_opt_process(record, saveFig = False, filename=None):

    
    if filename is None and saveFig:
        raise ValueError("Provide a filename in the form of '/path_to_save_the_file/name_of_the_file' without suffix!")

    fig, axs =plt.subplots(2,2,figsize=(9,6),dpi=128)
    
    plt.subplot(2,2,1)
    term = record['LL']
    plt.scatter(range(len(term)-2), term[2:], s=1)
    plt.xlabel('iteration')
    plt.ylabel('logLikelihood')
    plt.title('logLikelihood',fontsize=14)
    
    plt.subplot(2,2,2)
    term = record['alpha_norm']
    plt.scatter(range(len(term)), term, s=1)
    plt.xlabel('iteration')
    plt.ylabel('alpha l2-norm')
    plt.title('alpha l2-norm',fontsize=14)
    
    plt.subplot(2,2,3)
    term = record['delta_alpha']
    plt.scatter(range(len(term)), term, s=1)
    plt.xlabel('iteration')
    plt.ylabel('delta alpha')
    plt.title('delta alpha',fontsize=14)
    
    plt.subplot(2,2,4)
    term = [sum(x) for x in record['alpha']]
    plt.scatter(range(len(term)), term, s=1)
    plt.xlabel('iteration')
    plt.ylabel('sum alpha')
    plt.title('sum alpha',fontsize=14)
    
    fig.tight_layout()
    
    if saveFig:
        fig_title = filename.split('/')[-1]
        tit = fig.suptitle(fig_title, fontsize=16)
        fig.subplots_adjust(top=0.85)
        plt.savefig(filename+'.jpg',bbox_extra_artists=(tit,), bbox_inches='tight')

    plt.show()
'''
import math

def logRatio_dist(r_list,nbins=20,return_fig=None,rm_outlier=True,fig_size=(4,4),dpi=64):
    
    #output = para.get('output',None)
    custom_palette = sns.color_palette("Greens",5)
    
    x = np.log2(r_list)
    
    xy_lim = [math.floor(np.min(x)), math.ceil(np.max(x))] #[np.log2(1)-6, np.log2(1)+6] #[np.floor(v_min), np.floor(v_max)]
    x_m = round(np.mean(x))
    x_margin = np.max(np.array(xy_lim)-x_m)
    #xy_ticks = np.arange(xy_lim[0]+2, xy_lim[1], step=2)
    xy_ticks = np.arange(x_m-x_margin, x_m+x_margin+1, step=1)
    xy_ticks = xy_ticks[xy_ticks>0]
    
    if rm_outlier:
        x_shrinkage = rm_outliers(x)
    else:
        x_shrinkage = x
        
    #plt.figure(figsize=(6,8),dpi=256)
    fig, (ax_box,ax_hist) = plt.subplots(2, sharex=True, dpi=dpi,gridspec_kw={"height_ratios": (.1, .9)})
    fig.set_figheight(fig_size[0])
    fig.set_figwidth(fig_size[1])
    sns.boxplot(data=x, ax=ax_box, orient="h",fliersize=1,color = custom_palette[4])
    ax_box.set(xlabel='')

    xx = np.linspace(np.min(x)-1,np.max(x)+1,100)
    yy_shrinkage = scipy.stats.norm.pdf(xx,np.mean(x_shrinkage),np.std(x_shrinkage))

    a = plt.hist(x,nbins,label='trueR',alpha=0.5, color=custom_palette[4],density=True)
    top = np.max(a[0])*1.05
    plt.plot([np.mean(x_shrinkage), np.mean(x_shrinkage)], [0,top], linestyle='dashed', color='black')
    plt.plot(xx,yy_shrinkage,label='rm outliers',color='black')
    plt.xlabel('Total mRNA ratio',fontsize=18)
    plt.ylabel('Density',fontsize=18)
    plt.xticks(xy_ticks, np.round(2**xy_ticks,2), fontsize=12)
    #plt.yticks(np.arange(0,top+5,50),np.arange(0,top+5,50))
    #plt.suptitle(dlabel+', '+r'$e^\mu_r$='+str(round(2**np.mean(x_shrinkage),3)),fontsize=18)
    plt.suptitle(r'$R_{est}$='+str(round(2**np.mean(x_shrinkage),3)),fontsize=18)
    plt.tight_layout()
    if return_fig is True:
        return fig
    plt.show()




sns.set(style="ticks")

def get_bins(x,bw=0.05):
    return np.arange(min(x),max(x)+bw,bw)

def UMI_hist(adata,groupby=None, groups='all',return_fig=None,**fig_para):
    '''
    Visualize the log-UMI-amount distribution.

    Parameters
    ----------
    adata : AnnData
        The (annotated) UMI count matrix of shape `n_obs` × `n_vars`.
        Rows correspond to droplets and columns to genes.
    groupby : str
        The key of the droplet categories stored in adata.obs. 
    groups : list of strings, optional
        Droplet categories to visualize logUMI distributions. It should be eithor 'all' or a list of cell type annotations specified in adata.obs[groupby].
        The default is 'all'.
    return_fig : bool, optional
        Return the matplotlib figure. The default is None.
    **fig_para : optional
         Parameters to configure the plot.

    Raises
    ------
    ValueError
        If 'groups' specifies categories not matching with adata.obs[groupby].

    Returns
    -------
    fig : matplotlib figure

    '''
    
    x_hist='log2_total_UMIs'
    if 'log2_total_UMIs' not in adata.obs:
        adata.obs['log2_total_UMIs'] = np.log2(np.ravel(adata.X.sum(1)))
    
    obsdata = adata.obs
    
    bw = fig_para.get('bw',0.2)
    fix_bins = fig_para.get('fix_bins',None)
    fix_yticks = fig_para.get('fix_yticks',None)
    palette = fig_para.get('palette','Set2')

    fig = plt.figure(figsize=(6,4),dpi=64)
    ax=plt.subplot()
    
    if fix_bins is None:
        bins=get_bins(obsdata[x_hist],bw)
    else:
        bins=fix_bins   
        
    if groupby is None:
        col_list = sns.color_palette(palette, 1)
        plt.hist(obsdata[x_hist],bins,alpha=0.6, color=col_list[0])
    else:
        if groups == 'all':
            groups = obsdata[groupby].unique()
        elif isinstance(groups,list) and len([1 for v in groups if v not in obsdata[groupby].unique()]):
            raise ValueError("'show_groups' contains values absent in 'groupby'.")
            
        col_list = sns.color_palette(palette, len(groups))
        v_list = [obsdata[x_hist][obsdata[groupby]==g] for g in groups]        
        for i,g in enumerate(groups):
            plt.hist(v_list[i],bins,label=g,alpha=0.6, color=col_list[i])

        plt.legend(loc='upper right',fontsize=12)
        
    # Hide the right and top spines
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    plt.xlabel('Log2 Total UMI counts',fontsize=15)
    plt.ylabel('Frequency',fontsize=15)
    plt.xticks(fontsize=12)
    plt.yticks(fontsize=12)
    if fix_yticks is not None:
        plt.yticks(fix_yticks,fix_yticks,fontsize=12)
    #plt.xticks(xticks[xidx], xannos[xidx])
    if return_fig is True:
        return fig


def corrected_UMI_hist(adata,groupby,groups,reference,logUMIby,ratios,return_fig=None):
    '''
    Compare raw log-UMI-amount distributions versus ratio-based corrected log-UMI-amount distributions.

    Parameters
    ----------
    adata : AnnData
        The (annotated) UMI count matrix of shape `n_obs` × `n_vars`.
        Rows correspond to droplets and columns to genes.
    groupby : str
        The key of the droplet categories stored in adata.obs. 
    groups : list, optional
        Droplet categories, e.g. ['Homo-ct1', 'Homo-ct2'] annotated in adata.obs[groupby] to display. 
        The default is None.
        We recommand putting the reference group in the first element.
    reference : str
        One droplet category annotated in adata.obs[groupby].
        Use the specified group as reference. Generally, this group has the smallest mRNA content. 
    logUMIby : str
        The key of logUMI values stroed in adata.obs. The default is 'log10UMIs'.
    ratios : list
        List of estimated total-mRNA ratios. Each element corresponds to a group specified in 'groups'. 
        We recommand setting the value of reference group to be 1.
    return_fig : bool, optional
        Return the matplotlib figure. The default is None.

    Returns
    -------
    g : matplotlib figure


    '''
    ba_info = get_corrected_logUMI(adata,groupby,groups,reference,logUMIby,ratios)
    sns.cubehelix_palette(2*len(groups), rot=-.4, light=.7)
    sns.set_theme(style="white", rc={"axes.facecolor": (0, 0, 0, 0)})

    pal_list = sns.cubehelix_palette(2*len(groups), start=1.4, rot=-.25, light=.7, dark=.4) + \
    sns.cubehelix_palette(2*len(groups), rot=-.4, light=.7)
    pal = [pal_list[i] for i in [2*ii for ii in range(2*len(groups))]] 

    g = sns.FacetGrid(ba_info, row=groupby, hue=groupby, aspect=7, height=.8, palette=pal)
    g.map(sns.kdeplot, logUMIby, bw_adjust=.7, 
          cut=4, clip_on=True, fill=True, alpha=0.8, linewidth=1.5)
    g.map(sns.kdeplot, logUMIby, bw_adjust=.7, 
          cut=4, clip_on=True, color="w", lw=2)
    g.map(plt.axhline, y=0, linewidth=2, linestyle="-", color=None, clip_on=False)
    g.map(label, groupby)
    g.fig.subplots_adjust(hspace=-0.5)
    #g.set(yticks=[], xlabel="", ylabel="",xlim=(None, 4), xticks=[3,3.5,4],title="")
    g.set(yticks=[], xlabel="", ylabel="",title="")#,xticks=xticks[xidx], xannos=xannos[xidx],title="")
    #g.set_xticklabels(xticks[xidx],xannos[xidx])
    #g.set_xticklabels(xannos[xidx])

    g.despine(bottom=True, left=True)
    #plt.savefig('./outcome/fig/UMIhist.pdf')
    plt.show()
    if return_fig is True:
        return g



def get_corrected_logUMI(adata, groupby, groups, reference,logUMIby, ratios):
    
    didx = [d for d in adata.obs_names if adata.obs.loc[d,groupby] in groups]
    #ori_data = pd.DataFrame(adata[adata.obs['danno']!='Hetero-dbl',:].obs)
    ori_data = pd.DataFrame(adata[didx,:].obs)
    ori_data[groupby] = ori_data[groupby].astype(str)
    corr_data = copy.deepcopy(ori_data)
    for i,v in enumerate(groups):
        logratio_obs = adata.uns['para_logUMI'].loc[v,'mean']-adata.uns['para_logUMI'].loc[reference,'mean']
        delta_logUMI = np.log10(ratios[i])-logratio_obs
        corr_data.loc[ori_data[groupby]==v,logUMIby] = ori_data.loc[ori_data[groupby]==v,logUMIby]+delta_logUMI#np.log10(ratio[i])
        corr_data.loc[ori_data[groupby]==v,groupby] = v+'_corrected' 
        
    corr_data[groupby] = pd.Categorical(corr_data[groupby],groups+[v+'_corrected' for v in groups])
    ori_data[groupby] = pd.Categorical(ori_data[groupby],groups)
    ba_info = pd.concat([ori_data,corr_data])

    return ba_info



def label(x, color, label):
    ax = plt.gca()
    ax.text(0, .1, label, fontweight="bold", color=color,
            ha="left", va="center", transform=ax.transAxes)




with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    # fxn()
    
    
    
def volcano_2DE(de_df, marker_list=None, marker_colors=None, return_fig=None):
    '''
    Plot the volcano plot of 2 DE results stored in 'de_df'.

    Parameters
    ----------
    de_df : pandas.DataFrame
        Table fusing two input DE resutls.
    markers_dn2up : list of str, optional
        Marker genes to highlight. The default is None.

    Returns
    -------
    None.

    '''
    fc_x = de_df['log2FC_gs']
    fc_y = de_df['log2FC_rc']
    genes_dn2up = de_df.index[de_df['levelchange']=='dn2up']
    genes_ns2up = de_df.index[de_df['levelchange']=='ns2up']
    genes_dn2ns =  de_df.index[de_df['levelchange']=='dn2ns']

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        
        pval_x = - np.log10(de_df['pval_adj_gs'])
        pval_y = - np.log10(de_df['pval_adj_rc'])

    pval_x[pval_x==np.inf] = -np.log10(1e-323)
    pval_y[pval_y==np.inf] = -np.log10(1e-323)

    fig, axs =plt.subplots(1,2,figsize=(8,5),dpi=64)
    plt.subplot(1,2, 1)
    plt.scatter(fc_x,pval_x,s=0.5,color='silver')
    
    if marker_list is not None:
        for idx in range(len(marker_list)):
            plt.scatter(fc_x[marker_list[idx]], pval_x[marker_list[idx]],color=marker_colors[idx],s=20)
            
    plt.plot([0,0],[0,323],c='black',linestyle='dashed')
    plt.plot([-6,6],[-np.log10(0.05),-np.log10(0.05)],c='black',linestyle='dashed')
    plt.xticks([-5,0,5],[-5,0,5],fontsize=15) 
    plt.yticks([0,100,200,300],[0,100,200,300],fontsize=15) 
    plt.title('Global scaling DE',fontsize=18)
    plt.ylabel(r'$-log_10(p.val)$',fontsize = 18)
    plt.xlabel(r'$log_2FC$',fontsize = 18)

    plt.subplot(1,2, 2)
    plt.scatter(fc_y,pval_y,s=0.5,color='silver')

    if marker_list is not None:
        for idx in range(len(marker_list)):
            plt.scatter(fc_y[marker_list[idx]], pval_y[marker_list[idx]],color=marker_colors[idx],s=20)
            
    plt.plot([0,0],[0,323],c='black',linestyle='dashed')
    plt.plot([-6,6],[-np.log10(0.05),-np.log10(0.05)],c='black',linestyle='dashed')
    plt.xticks([-5,0,5],[-5,0,5],fontsize=15) 
    plt.yticks([0,100,200,300],[0,100,200,300],fontsize=15) 
    plt.title('RC scaling DE',fontsize=18)
    #plt.xlabel(r'$-log_10(p.val)$',fontsize = 18)
    plt.xlabel(r'$log_2FC$',fontsize = 18)
    plt.tight_layout()
    if return_fig is True:
        return fig
    plt.show()




def get_plot_df(adata, genes):

    #exp_vec = np.ravel(scipy.sparse.csr_matrix.todense(adata.raw[:,genes].X), order = 'F')
    exp_vec = np.ravel(adata[:,genes].X.toarray(), order = 'F')
    data_df = pd.DataFrame({'barcodes':adata.obs_names.tolist()*len(genes),
                            'expression': exp_vec,
                            'genes': list(itertools.chain.from_iterable([[g]*adata.n_obs for g in genes])),
                            'celltype': adata.obs['danno'].tolist()*len(genes)})
    
    return data_df




def violin_2DE(adata_1, adata_2, genes, corrected='para', data_name=['Before correction','After correction'], return_fig=None):
    '''
    Violin plot of gene expressions in global-scaling values and ratio-corrected values.

    Parameters
    ----------
    adata_1 : AnnData
        The expression matrix after global-scaling normalization.
        Rows correspond to droplets and columns to genes.
    adata_2 : AnnData
        The expression matrix after total-mRNA-ratio-based correction.
        Rows correspond to droplets and columns to genes.
    genes : list of str
        Genes to plot.
    return_fig : bool, optional
        Return the matplotlib figure. The default is None.

    Returns
    -------
    g : matplotlib figure

    '''
    data_df1 = get_plot_df(adata_1, genes)
    if 'log1p' not in adata_1.uns:
        data_df1['expression'] = np.log1p(data_df1['expression'])
    
    data_df2 = get_plot_df(adata_2, genes)
    if 'log1p' not in adata_2.uns:    
        data_df2['expression'] = np.log1p(data_df2['expression'])
    
    if 'corrected' in adata_2.uns and adata_2.uns['corrected'] == 'para':
        # shift data when displaying expression
        data_df2['expression'] = data_df2['expression'] + np.log(adata_2.uns['shift_ratio'])
    
    data_df = pd.concat([data_df1, data_df2])
    data_df['DE'] = [data_name[0]]*data_df1.shape[0] + [data_name[1]]*data_df2.shape[0]
    
    g = sns.catplot(x="genes", y="expression", hue="celltype",
                    col="DE",
                    data=data_df, kind="violin",split=True,inner = 'quartile',
                    height=4, aspect=0.15*(1+len(genes)),cut=0,bw=0.3)

    if return_fig is True:
        return g
 
    
    
 
def violin(adata,genes,return_fig=True):
    '''
    Violin of gene expression.

    Parameters
    ----------
    adata : TYPE
        DESCRIPTION.
    genes : TYPE
        DESCRIPTION.
    return_fig : TYPE, optional
        DESCRIPTION. The default is True.

    Returns
    -------
    g : TYPE
        DESCRIPTION.

    '''
    data_df = get_plot_df(adata, genes)

    g = sns.catplot(x="genes", y="expression", hue="celltype",
                    data=data_df, kind="violin",split=True,inner = 'quartile',
                    height=4, aspect=0.18*(1+len(genes)),cut=0,bw=0.3)
    
    if return_fig is True:
        return g
 
     
        
    
def DEshift(df,return_fig=None):
    
    x1 = [[df.loc['up',:].sum(), df.loc['up','up']], 
          # ns -> up
          [df.loc['up',:].sum()+df.loc['ns',:].sum(), df.loc['up','up']+df.loc['ns','up']], 
          # dn -> up
          [df.loc['up','up'],df['up'].sum()], 
          # up -> ns
          [df.loc['up',:].sum()+df.loc['ns',:].sum()+df.loc['dn','up'],df['up'].sum()+df['ns'].sum()-df.loc['dn','ns']], 
          # dn -> ns
          [df.loc['up','up']+df.loc['up','ns'], df['up'].sum()+df['ns'].sum()], 
          # up -> dn
          [df.loc['up',:].sum()+df.loc['ns','up']+df.loc['ns','ns'], df['up'].sum()+df['ns'].sum()+df.loc['up','dn']] 
          # ns -> dn
         ]
    x2 = [[df.loc['up',:].sum()+df.loc['ns','up'], df.loc['up','up']+df.loc['ns','up']],# ns -> up
          [df.loc['up',:].sum()+df.loc['ns',:].sum()+df.loc['dn','up'], df['up'].sum()],# dn -> up
          [df.loc['up','up']+df.loc['up','ns'],df['up'].sum()+df.loc['up','ns']],# up -> ns
          [df.sum().sum()-df.loc['dn','dn'],df['up'].sum()+df['ns'].sum()],# dn -> ns
          [df.loc['up',:].sum(),df['up'].sum()+df['ns'].sum()+df.loc['up','dn']],# up -> dn
          [df.loc['up',:].sum()+df.loc['ns',:].sum(),df.sum().sum()-df.loc['dn','dn']]# ns -> dn
         ]

    shading_col = ['blue','red','green','#fac748','#f88dad','#1f7a8c'] # color of DE change
    deltaDE_list = ['ns2up','dn2up','up2ns','dn2ns','up2dn','ns2dn']

    dataset = [df.sum(1).to_dict(), df.sum(0).to_dict()]
    data_orders = [['up','ns','dn'],['up','ns','dn']]

    #colors = ["#161a1d","grey","#a4161a"] # color of DE, dn ns up
    colors = ["blue","grey","red"] # color of DE, dn ns up

    names = sorted(dataset[0].keys())
    values = np.array([[data[name] for name in order] for data,order in zip(dataset, data_orders)])
    lefts = np.insert(np.cumsum(values, axis=1),0,0, axis=1)[:, :-1]
    orders = np.array(data_orders)
    bottoms = np.arange(len(data_orders))

    fig,ax = plt.subplots(figsize=(15,3))
    for name, color in zip(names, colors):
        idx = np.where(orders == name)
        value = values[idx]
        left = lefts[idx]
        plt.bar(x=left, 
                height=0.3, 
                width=value, 
                bottom=bottoms, 
                color=color, 
                alpha=0.6,
                orientation="horizontal")#, 
                #label=name)

    for i in range(6):
        if sum(np.array(x2[i]-np.array(x1[i]))) > 0:
            ax.fill_betweenx(y =[0.18,0.82], x1 = x1[i], x2 = x2[i], color = shading_col[i], label=deltaDE_list[i])
    
    up,ns,dn = df.sum(0)
    plt.text(up/2-300,0.92,up,fontsize=25)
    plt.text((up+up+ns)/2-300,0.92,ns,fontsize=25)
    plt.text((2*up+2*ns+dn)/2-300,0.92,dn,fontsize=25)

    up,ns,dn = df.sum(1)
    plt.text(up/2-300,-0.08,up,fontsize=25)
    plt.text((up+up+ns)/2-300,-0.08,ns,fontsize=25)
    plt.text((2*up+2*ns+dn)/2-300,-0.08,dn,fontsize=25)
    
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    #plt.yticks(bottoms+0.4, ["data %d" % (t+1) for t in bottoms])
    plt.yticks([])
    #plt.xticks([])
    plt.legend(loc="best", bbox_to_anchor=(1.0, 1))
    plt.subplots_adjust(right=0.85)

    plt.tight_layout()
    if return_fig is True:
        return fig
    plt.show()







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr  6 15:04:27 2021

@author: qy
"""

import numpy as np
import scipy
from scipy.special import gammaln
#import time
#import pickle
import traceback
import pandas as pd
#import os, sys
import pyDIMM
from scipy import stats
import multiprocessing as mp
import anndata

#%% DMN optimization 

def dmn(adata, groupby, groups='all', tol=1e-3, maxiter=1000, subset=None, verbose=2, verbose_interval=10):
    '''
    Fit Dirichlet-Multinomial distribution with UMI counts of homo-droplet populations.

    Parameters
    ----------
    adata : AnnData
        The (annotated) UMI count matrix of shape `n_obs` × `n_vars`.
        Rows correspond to droplets and columns to genes.
    groupby : str
        The key of the droplet categories stored in adata.obs. 
    groups : list of strings, optional
        Droplet categories to which DMN model shoudl be fitted with. It should be eithor 'all' or a list of cell type annotations specified in adata.obs[groupby].
        The default is 'all'.
    tol : float, optional
        The convergence threshold. The iteration will stop when the likelihood gain is below this threshold. The default is 1e-3.
    maxiter : int, optional
        The maximal number of iteration. The default is 1000.
    subset : int, optional
        Number of downsampled droplets to fit DMN. It is only recommended when you have excessive droplet numbers for homotypic droplet populations. The default is None.
    verbose : int, optional
        Enable verbose output. If 1 then it prints the current initialization and each iteration step. If greater than 1 then it prints also the log probability and time needed for each step.
        The default is 2.
    verbose_interval : int, optional
        Number of iteration done before the next print. The default is 10.

    Returns
    -------
    None.
    The optimized alpha vectors for each kind of droplet population is stroed in adata.varm['para_diri'].

    '''
    
    if groups=='all':
        groups = [d for d in adata.obs[groupby].unique() if len(d.split('_'))==1]
        
    counts_list = []
    vidx_list = []
    for group in groups:
    
        counts = adata[adata.obs[groupby]==group,:].X
        if isinstance(counts,scipy.sparse.csr.csr_matrix) or isinstance(counts, anndata._core.views.ArrayView):
            counts = counts.toarray()
        
        if not np.issubdtype(counts.dtype, np.integer):
            counts = counts.astype(int)
        
        vidx = counts.sum(0)!=0
        vidx_list.append(vidx)
        counts = counts[:,vidx]
        
        if subset is not None:
            if counts.shape[0] > subset:
                idx = np.random.choice(counts.shape[0], subset, replace=False)
                counts = counts[idx,:]
                
        counts_list.append(counts)
        
    n_cores = min(len(groups), int(mp.cpu_count()*0.8))
    
    pool = mp.Pool(n_cores)  
    result_compact = [pool.apply_async( _fitdmn, (counts_list[i], groups[i], tol,maxiter, verbose, verbose_interval) ) for i in range(len(groups))]
    pool.close()
    pool.join()
    
    alpha_out = [term.get() for term in result_compact] 
    
    #alpha_df = pd.DataFrame(np.array(alpha_out).T, index=adata.var_names, columns=groups)
    alpha_df = pd.DataFrame(np.ones([adata.n_vars,len(groups)])*1e-12, index=adata.var_names, columns=groups)
    for gidx,gval in enumerate(groups):
        alpha_df.loc[vidx_list[gidx],gval] = alpha_out[gidx]
    
    if 'para_diri' in adata.varm:
        for g in groups:
            adata.varm['para_diri'][g] = alpha_df[g]
    else:
        adata.varm['para_diri'] = alpha_df
    
    


def _fitdmn(counts,group,tol,maxiter,verbose,verbose_interval):
    
    dmm = pyDIMM.DirichletMultinomialMixture(n_components=1, 
                                             max_iter=maxiter, 
                                             n_init=1, 
                                             tol=tol,
                                             verbose=verbose,
                                             verbose_interval=verbose_interval).fit(counts)
    print(group+' is done!')
    return np.ravel(dmm.alphas)






#%% logNormal distribution optimization

def rm_outliers(x):
    iqr = stats.iqr(x)
    outlier_lb = np.quantile(x,0.25)-1.5*iqr
    outlier_ub = np.quantile(x,0.75)+1.5*iqr
    x_shrinkage = x[x > outlier_lb]
    x_shrinkage = x_shrinkage[x_shrinkage<outlier_ub]
    return x_shrinkage#,(outlier_lb,outlier_ub)


def logNormal(adata,groupby,groups='all',inplace=True):
    '''
    Fit logNormal distributions with UMI amounts of homo-droplet populations.

    Parameters
    ----------
    adata : AnnData
        The (annotated) UMI count matrix of shape `n_obs` × `n_vars`.
        Rows correspond to droplets and columns to genes.
    groupby : str
        The key of the droplet categories stored in adata.obs. 
    groups : list of strings, optional
        Droplet categories to fit the total UMIs wtih logNormal distribution. It should be eithor 'all' or a list of cell type annotations specified in adata.obs[groupby].
        The default is 'all'.
    inplace : bool, optional
        If or not to store fitted parameters into adata. The default is True.

    Returns
    -------
    para : list
        Retuen mean and std of fitted logNormal distributions if inplace is False.
    '''
    if groups=='all':
        groups = adata.obs[groupby].unique()
    
    if 'total_UMIs' not in adata.obs:
        adata.obs['total_UMIs'] = np.ravel(adata.X.sum(1))
    
    if 'log2_total_UMIs' not in adata.obs:
        adata.obs['log2_total_UMIs'] = np.log2(adata.obs['total_UMIs'])        

    para = []
    for g in groups:
        m,s = stats.norm.fit(rm_outliers(adata.obs['log2_total_UMIs'][adata.obs[groupby]==g]))
        para.append([m,s])

    if inplace:
        adata.uns['para_logUMI'] = pd.DataFrame(np.array(para),
                                                       index=groups,
                                                       columns=['mean','std'])
    else:
        return para







'''

class HiddenPrints:
    # Hide prints from pyDIMM in C
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def _job_fitdmn(counts, group,c_version,maxiter,output):

    print('Start fitting '+group+' droplets. This may take a long time. Please wait...\n')
    if c_version:
        t0 = time.time()
        with HiddenPrints():
            dimm = pyDIMM.DIMM(observe_data=counts, n_components=1, print_log=True)
            dimm.EM(max_iter=maxiter, max_alpha_tol=1e-3, max_loglik_tol=1e-3, save_log=True)
        print('Finish fitting '+group+' droplets. Time cost:', time.time()-t0, 'seconds.')

        log = pd.read_csv('pyDIMM.log')
        log.to_csv(os.path.join(output,group+'.dmnlog.txt'))
        #os.remove('pyDIMM.log')
        alpha_vec = np.ravel(dimm.get_model()['alpha'])

    else:

        log = open(os.path.join(output,group+'.dmnlog.txt'),'w',buffering=1)    
        log.write('niter'+','+'alpha_sum'+','+'loglik'+','+'alpha_l2norm'+','+'alpha_l2norm_delta'+'\n')

        t0 = time.time()
        alpha_init = initialize_alpha(counts)
        alpha_vec = optimize_alpha(counts, alpha_init, log=log,maxiter=maxiter)
        print('Finish fitting '+group+' droplets. Time cost:', time.time()-t0, 'seconds.\n')
        log.close()
    
    return alpha_vec

'''

def initialize_alpha(Y):

    gene_tot = np.ravel(np.sum(Y, axis=0))
    lib_size = np.ravel(np.sum(Y, axis=1))
    
    G = sum(gene_tot > 0)
    #C = sum(lib_size > 0)
    
    new_Y = Y[lib_size > 0, :]
    new_Y = new_Y[:, gene_tot > 0] 
    
    p_mat = ( new_Y.T/lib_size[lib_size > 0] ).T
    p_mean = p_mat.mean(0)
    p_var = p_mat.var(0)
    p_var[p_var == 0] = np.nanmean(p_var)
    
    alpha_sum = np.exp( sum(np.log(p_mean[:-1]*(1-p_mean[:-1])/p_var[:-1] - 1))/(G-1) )
    
    alpha = np.array([0.0]*Y.shape[1])
    alpha[gene_tot > 0] = alpha_sum * p_mean
    alpha[alpha == 0.0] = 1e-6

    return alpha
    


def optimize_alpha(Y, alpha, **para):
    
    minAlpha = 1e-100
    
    maxiter = para.get('maxiter',2000)
    likTol = para.get('likTol',1e-2)
    log = para.get('log',None)
    delta_alpha_Tol = para.get('alpTol',1e-1)
    keeptimes =  para.get('ntimes',100)
    
    delta_LL = 100.0
    delta_alpha = 20
    iteration = 0
    
    lib_size = np.ravel(np.sum(Y, axis=1))
    C,G = Y.shape
    
    #record = {'alpha': [], 'LL': [], 'alpha_norm': [], 'delta_alpha':[]}
    
    logLik = sum(calculate_LL(alpha, Y))
    alpha_norm = np.linalg.norm(alpha)
    
    #record['alpha'].append(alpha)
    #record['LL'].append(logLik)
    #record['alpha_norm'].append(alpha_norm)
    
    alpha_hits = []
    
    try:
    
        while( ((delta_LL > likTol) or len(alpha_hits) <  keeptimes) and (iteration < maxiter)):
            
            # calculate new alpha
            alphaNew = np.zeros(alpha.shape)
            den = sum(lib_size/(lib_size -1 + alpha.sum() ))
            for gidx in range(G):
                dataG = (( Y[:,gidx] + minAlpha)/(Y[:,gidx] + minAlpha - 1 + alpha[gidx]))
                num = dataG.sum()
                alphaNew[gidx] = alpha[gidx] * num / den
            
            # calculate new logLik
            LL_tmp = calculate_LL(alphaNew, Y)
            newLogLik  = LL_tmp.sum()
            
            # calculate diff to check convergence
            delta_LL = np.abs( (newLogLik - logLik)/logLik*100.0 )
            delta_alpha = np.linalg.norm(alphaNew - alpha)
            # update 
            alpha = alphaNew
            alpha[alpha <= 0] = minAlpha
            logLik = newLogLik
            
            alpha_norm = np.linalg.norm(alpha)
            
            #record['alpha'].append(alpha)
            #record['LL'].append(logLik)
            #record['alpha_norm'].append(alpha_norm)
            #record['delta_alpha'].append(delta_alpha)
            iteration += 1
            #if log is None:
            #    print('iter '+str(iteration)+': precision = '+str(sum(alpha)) + ', LL = ' + str(logLik))
            #else:
            log.write(str(iteration)+','+str(sum(alpha)) +','+str(logLik) +','+str(alpha_norm)+','+str(delta_alpha)+ '\n')
            
            # if delta_alpha is less than delta_alpha_Tol in continuous keeptimes, then break
            alpha_hits.append( delta_alpha < delta_alpha_Tol)
            if not np.all(alpha_hits):
                alpha_hits = []
        #elif len(alpha_hits) >= keeptimes:
        '''
        if log is None:
            print('terminated after '+str(iteration)+' iterations.')
        else:
            log.write('terminated after '+str(iteration)+' iterations.\n')
            #break        
        '''
    except Exception as e:
        #if log is None:
        print('repr(e):\n',repr(e))
        print('traceback.print_exc():')
        traceback.print_exc()
        '''
        else:
            log.write('repr(e):\t'+repr(e)+'\n')
            log.write('traceback.print_exc():\n'+str(traceback.print_exc())+'\n')
        '''
    return alpha
    



def lgammaVec(inVec):
    outVec = gammaln(inVec)
    idx_inf = np.isinf(outVec)
    outVec[idx_inf] = 709.1962086421661
    # lnGamma(1e-308) = 709.1962086421661, lnGamma(1e-309) = inf
    # if val is closer than 1e-308 to 0, e.g. 1e-309, inf will be returned 
    # we would truncate it to a certain number: here it's lnGamma(1e-308)
    return outVec




def calculate_LL(alpha, Y):
    
    lib_size = np.ravel(np.sum(Y, axis=1))
    C = Y.shape[0]
    lgammaAlpha = lgammaVec(alpha)
    alphaSum = alpha.sum()    
    
    LL_tmp = np.array([0]*C)
    num1 = np.array([0]*C)#np.zeros((C,1))
    num2 = np.array([0]*C)#np.zeros((C,1))
    for i in range(C):
        dataAlphaTmp = Y[i,:] + alpha
        num1[i] = sum(lgammaVec(dataAlphaTmp) - lgammaAlpha)
        num2[i] = gammaln(alpha.sum()) - gammaln( alphaSum + lib_size[i] )
        LL_tmp[i] = num1[i] + num2[i] 
    
    return LL_tmp





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul  6 02:41:49 2022

@author: qy
"""

#import pickle
import numpy as np
import scipy
from scipy.special import gammaln, digamma
import os
import pandas as pd
import scanpy as sc
import tqdm
import multiprocessing as mp
import math
from scipy import stats




import gzip
import shutil
from scipy import sparse, io

def write_h5ad_to_mtx(adata, droplets_to_write, features_to_write, outfolder):

    adata = adata[droplets_to_write, features_to_write].copy()
    sparse_matrix = sparse.csr_matrix(adata.X.transpose()) 

    os.makedirs(outfolder, exist_ok=True)
    io.mmwrite(os.path.join(outfolder,'matrix.mtx'),sparse_matrix)
    with gzip.open(os.path.join(outfolder,'barcodes.tsv.gz'), 'wb') as barcode_file:
        for barcode in droplets_to_write:
            barcode_file.write('{}\n'.format(barcode).encode())
    with gzip.open(os.path.join(outfolder,'features.tsv.gz'), 'wb') as feature_file:
        for feature in features_to_write:
            feature_file.write('{}\n'.format(feature).encode())
    with open(os.path.join(outfolder,'matrix.mtx'),'rb') as mtx_in:
        with gzip.open(os.path.join(outfolder,'matrix.mtx') + '.gz','wb') as mtx_gz:
            shutil.copyfileobj(mtx_in, mtx_gz)
    os.remove(os.path.join(outfolder,'matrix.mtx')) 
    



def annote_clusters(adata_psgl, ct_mapper, groupby='leiden',anno_group=None):
    '''
    Annotate clusters with domain knowledge of cell  type markers.

    Parameters
    ----------
    adata_psgl : AnnData
        UMI matrix of putative singlets (homotypic droplets).
    ct_mapper : dic
        Domain knowledge of cell type markers. The format must be {'gene1':'celltype1','gene2':'cellltype2','gene3':'celltype3'}.
    groupby : 'str'
        The key of the clustering results stored in adata.obs. The default is 'leiden'.
    anno_group : 'str', optional
        The key of the output annotation of clusters stored in adata.obs. The default is value of 'groupby' with '_anno' as suffix.

    Returns
    -------
    None. The results are saved in adata.obs["groupby"_anno'].

    '''

    if 'rank_genes_groups' not in adata_psgl.uns:
        sc.tl.rank_genes_groups(adata_psgl, groupby, method='wilcoxon')

    result = adata_psgl.uns['rank_genes_groups']
    groups = result['names'].dtype.names

    # ct_mapper = {'g_2426':'A','g_10031':'B','g_5031':'C','g_5497':'D'} 

    # Extract the logFC of the input marker genes in all clusters 
    lf_list = []
    for k in groups:

        gidx = [list(result['names'][k]).index(g) for g in ct_mapper.keys()]
        lf_list.append(result['logfoldchanges'][k][gidx])

    lf_df = pd.DataFrame(np.array(lf_list),index=groups,columns=ct_mapper.keys())
    # row: cluter, col: marker, value: logFC

    clusters = [lf_df.index[np.argmax(lf_df[g])] for g in ct_mapper.keys()]
    annos = dict(zip(clusters, ct_mapper.values()))
    # map clusters to cell types 

    if anno_group is None:
        anno_group = groupby + '_anno'

    adata_psgl.obs[anno_group] = [annos[g] for g in adata_psgl.obs[groupby]]

    print('Cell type annotation is saved in .obs['+anno_group+'].')




def extract_specific_genes(adata_psgl, groupby, pval=0.001, logfc=0):
    '''
    Extract specifically highly expressed genes of each cell type based on DE results.

    Parameters
    ----------
    adata_psgl : AnnData
        UMI matrix of putative singlets (homotypic droplets).
    groupby : 'str'
        The key of the cell type annotation results stored in adata.obs.
    pval : float, optional
        Cutoff of p value to extract significant DE genes. The default is 0.001.
    logfc : float, optional
        Cutoff of log2Fold-change to extract cell-type-specific DE genes. The default is 0.

    Returns
    -------
    degene_sorted : list
        Identified heterotypic doublets composed by the input cell type pair.

    '''
    
    if 'rank_genes_groups' not in adata_psgl.uns:
        sc.tl.rank_genes_groups(adata_psgl, groupby, method='wilcoxon')
        
    result = adata_psgl.uns['rank_genes_groups']
    groups = result['names'].dtype.names
    de_df = pd.DataFrame({group + '_' + key[:1]: result[key][group]
                          for group in groups for key in ['names', 'pvals','logfoldchanges']})

    topn = np.min([len([i for i in de_df.index if de_df.loc[i,ct+'_p']<pval and de_df.loc[i,ct+'_l']>logfc]) for ct in groups])
    
    gtmp_list = []
    for n,v in enumerate(groups):
        
        ridx = [gidx for gidx,gname in enumerate(de_df[v+'_n']) if de_df.iloc[gidx,3*n+1]<pval and de_df.iloc[gidx,3*n+2]>0]
        gtmp = de_df.iloc[ridx,n*3:(n+1)*3]
        
        gtmp = gtmp.sort_values(by=[v+'_l'],ascending=False)
        gtmp = gtmp.reset_index(drop=True)
        gtmp_list.append(gtmp.iloc[:topn,:])

    degene_sorted = pd.concat(gtmp_list, 1)
    
    return degene_sorted





#%% meta-genes

def cal_KL_bc(adata, groups, parallel=True):
    '''
    Calculate the KL-divergence of genes betweeo two Dirichlet-Multinomial distribution.

    Parameters
    ----------
    adata : AnnData
        The (annotated) UMI count matrix of shape `n_obs` × `n_vars`.
        Rows correspond to droplets and columns to genes.
    groups : list of strings
        Two droplet categories, e.g. ['Homo-ct1', 'Homo-ct2'] annotated in adata.obs[groupby] to which KL-divergence should be calculated with.
    parallel : bool, optional
        If or not to run in parallel. The default is True.

    Returns
    -------
    None.

    '''

    raw_Alpha_1 = adata.varm['para_diri'][groups[0]]
    raw_Alpha_2 = adata.varm['para_diri'][groups[1]]
    
    if parallel:
        
        tot_objs = len(raw_Alpha_1)
        
        num_cores = int(mp.cpu_count()*0.8)
        job_unit = math.floor(tot_objs/num_cores)
        
        num_jobs = tot_objs // job_unit
        jobs_sidx = [i*job_unit for i in range(num_jobs)]
        jobs_sidx.append(tot_objs)
        
        pool = mp.Pool(num_cores)  
        result_compact = [pool.apply_async( job_KL, (raw_Alpha_1,raw_Alpha_2, jobs_sidx[i], jobs_sidx[i+1]) ) for i in range(num_jobs)]
        pool.close()
        pool.join()        
        
        kl = np.array([v for term in result_compact for v in term.get()])
        
    else:
        
        # KL-metric for each gene
        kl_1_2 = []
        kl_2_1 = []
        
        # add progression bar if not multiprocess
        for gidx in tqdm.tqdm(range(len(raw_Alpha_1))):
            
            kl_tmp = KL_Dir_miginal(raw_Alpha_1, raw_Alpha_2, gidx) 
            kl_1_2.append(kl_tmp)
            
            kl_tmp = KL_Dir_miginal(raw_Alpha_2, raw_Alpha_1, gidx) 
            kl_2_1.append(kl_tmp)
        
        kl = (np.array(kl_1_2)+np.array(kl_2_1))/2
    
    '''
    info = {'raw_alpha1':raw_Alpha_1,
            'raw_alpha2':raw_Alpha_2,
            'kl':kl}
    
    f = open(os.path.join(output,'dmn.kl.pickle'),'wb')
    pickle.dump(info,f)
    f.close()
    '''
    #kl = obtain_KL_given2alphaVec(raw_Alpha_1, raw_Alpha_2, output)
    adata.varm['kl'] = pd.DataFrame({groups[0]+'_'+groups[1]:kl},index=adata.var_names)    
    
    #return kl
    
'''
def obtain_mg(raw_Alpha1_0, raw_Alpha2_0, kl, kl_cutoff=1, merging_threshold=5, skip_threshold=2,alphaMin = 1):

    # use genes with KL greater than cutoff to obtain meta-genes
    KL_filter_idx = np.where(kl > kl_cutoff)[0]
    
    raw_Alpha1 = raw_Alpha1_0[KL_filter_idx]
    raw_Alpha2 = raw_Alpha2_0[KL_filter_idx]

    mg_dic = get_mg(raw_Alpha1, raw_Alpha2, merging_threshold, skip_threshold,alphaMin)
    
    mg_pool = mg_dic['mg_pool'] 
    #balanced_mg_num = mg_dic['balanced_mg_num']
    
    # original idx of genes used to generate metagenes, each element is a list of gene idx
    mg_dic['mg_gidx'] = [list(KL_filter_idx[term]) for term in mg_pool]
    mg_dic['mg_genepool'] = [g for sub in  mg_dic['mg_gidx']  for g in sub]
    
    return mg_dic
'''




def job_KL(raw_Alpha_1,raw_Alpha_2,start_sidx, end_sidx):
    
    sub_kl = []
    for gidx in range(start_sidx,end_sidx):

        kl_1_2 = KL_Dir_miginal(raw_Alpha_1, raw_Alpha_2, gidx) 
        kl_2_1 = KL_Dir_miginal(raw_Alpha_2, raw_Alpha_1, gidx) 
        
        kl = (kl_1_2+kl_2_1)/2
        sub_kl.append(kl)
        
    return sub_kl



    
def KL_Dir_miginal(alpha1, alpha2, gidx):
    
    # gammaln(1e-309) = inf, gammaln(1e-308) = 709.1962086421661
    alpha1[alpha1 < 1e-6] = 1e-6
    alpha2[alpha2 < 1e-6] = 1e-6
    
    #print(gidx)
    g_alpha1 = alpha1[gidx]
    g_alpha2 = alpha2[gidx]
    left_alpha1 = sum(alpha1) - g_alpha1
    left_alpha2 = sum(alpha2) - g_alpha2
    kl = gammaln(sum(alpha1)) - gammaln(g_alpha1) - gammaln(left_alpha1) - gammaln(sum(alpha2)) + gammaln(g_alpha2) + gammaln(left_alpha2) + \
         (g_alpha1 - g_alpha2)*(digamma(g_alpha1)-digamma(sum(alpha1))) + (left_alpha1 - left_alpha2)*(digamma(left_alpha1) - digamma(sum(alpha1)))

    return kl



#%% correct UMI 

import copy

def correctUMI(adata, groupby, method='upsampling', logUMIby=None):
    '''
    Correct UMI to meet the estimated total-mRNA-ratio.

    Parameters
    ----------
    adata : AnnData
        The (annotated) UMI count matrix of shape `n_obs` × `n_vars`.
        Rows correspond to droplets and columns to genes.
    groupby : str
        The key of the droplet categories stored in adata.obs. 
    logUMIby : str, optional
        The key of total UMIs in log2 stored in adata.obs. The default is None.
        
    Returns
    -------
    adata_rc : TYPE
        The corrected UMI count matrix.

    '''    

    
    #groups = list(ratios.keys())
    #r_vals = list(ratios.values()) 
    #resortidx = np.argsort(r_vals)
    
    #groups = [groups[v] for v in resortidx]
    #r_vals = [r_vals[v] for v in resortidx]

    groups = adata.uns['ratio_serial'].index.values.tolist()
    r_vals = np.ravel(adata.uns['ratio_serial'].values)

    alpha_df = adata.varm['para_diri']

    if logUMIby is None:
        if 'log2_total_UMIs' not in adata.obs:
            adata.obs['total_UMIs'] = np.ravel(adata.X.sum(1))
            adata.obs['log2_total_UMIs'] = np.log2(adata.obs['total_UMIs']) 
            
        logUMIby = 'log2_total_UMIs'

        
    #logN_para = [norm.fit(adata.obs[logUMIby][adata.obs[groupby]==v]) for v in groups]
    #mu_list = [x[0] for x in logN_para]
    #std_list = [x[1] for x in logN_para]
    mu_list = list(adata.uns['para_logUMI'].loc[groups,'mean'])
    
    mat_rc = copy.deepcopy(adata.X.toarray())
    
    for i in range(len(groups)-1):
        
        print('Correct UMIs of population '+groups[i+1])

        UMI_delta = 2**(mu_list[0] + np.log2(r_vals[i+1]) - mu_list[i+1])-1
        alpha = alpha_df[groups[i+1]]
        X_mat = copy.deepcopy(mat_rc[adata.obs[groupby]==groups[i+1],:])
        
        X_mat_up = _correct(X_mat, alpha, UMI_delta, method)
        mat_rc[adata.obs[groupby]==groups[i+1],:] = X_mat_up
    
    adata_rc = sc.AnnData(mat_rc)
    #adata_rc.obs = adata.obs
    adata_rc.var_names = adata.var_names.values.tolist()
    adata_rc.obs_names = adata.obs_names.values.tolist()
    
    for col in adata.obs:
        adata_rc.obs[col] = adata.obs[col].values
    adata_rc.varm['para_diri'] = adata.varm['para_diri']
    adata_rc.obs['rc_total_UMIs'] = np.ravel(adata_rc.X.sum(1))
    adata_rc.obs['log2_total_UMIs'] = np.log2(adata_rc.obs['rc_total_UMIs'])
    
    for term in adata.uns:
        adata_rc.uns[term] = adata.uns[term]
        
    adata_rc.uns['corrected'] = 'data'

    return adata_rc




def _correct(X_mat, alpha, UMI_delta, method):

    if isinstance(X_mat, scipy.sparse.csr.csr_matrix):
        X_mat = X_mat.toarray()
    
    if method == 'upscaling':
        
        X_mat_up = X_mat * (1+UMI_delta)
        X_mat_up = X_mat_up.astype(int)
        
    elif method == 'upsampling':
        x_list = []
        for cidx in tqdm.tqdm(range(X_mat.shape[0])):
            x = X_mat[cidx]
            p = np.random.dirichlet(alpha+x,size=1)
            x_delta = np.random.multinomial(int(sum(x)*UMI_delta), p[0])
            x_modified = x + x_delta
            x_list.append(x_modified)
    
        X_mat_up = np.array(x_list)
        #X_mat_up_sparse = scipy.sparse.csr_matrix(X_mat_up)
    else:
        raise ValueError("'method' value error: only 'upsampling' or 'upscaling' is supported.")

    return X_mat_up #_sparse




def rm_outliers(x):
    iqr = stats.iqr(x)
    outlier_lb = np.quantile(x,0.25)-1.5*iqr
    outlier_ub = np.quantile(x,0.75)+1.5*iqr
    x_shrinkage = x[x > outlier_lb]
    x_shrinkage = x_shrinkage[x_shrinkage<outlier_ub]
    return x_shrinkage#,(outlier_lb,outlier_ub)



from scipy.stats import norm
 
def correct_para(adata_sgl, groupby, achor_group=None, adjust_group=None):
    
    mean_logR, std_logR = norm.fit(rm_outliers(np.log10(adata_sgl.uns['Rest_perdbl'])))
    
    adata_sgl.uns['raw_para_logUMI'] = copy.deepcopy(adata_sgl.uns['para_logUMI'])
    adata_sgl.uns['corrected'] = 'para'
    
    if achor_group is None or adjust_group is None:
        groups = list(adata_sgl.obs[groupby].unique())
        adjust_group,achor_group = (np.argsort(adata_sgl.uns['para_logUMI'].loc[groups,'mean'])).index
        # adjust the parameter of group with smaller mean by default
        
    achor_mean = adata_sgl.uns['para_logUMI'].loc[achor_group,'mean']
    adjust_mean = adata_sgl.uns['para_logUMI'].loc[adjust_group,'mean']
    
    adata_sgl.uns['shift_ratio'] = 2**(mean_logR - np.abs(achor_mean-adjust_mean))
    
    if achor_mean > adjust_mean:
        mean_corrected = achor_mean - mean_logR
    else:
        mean_corrected = achor_mean + std_logR
        
    adata_sgl.uns['para_logUMI'].loc[adjust_group,'mean'] = mean_corrected

    






import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="tomas",
    version="0.1.21",
    license='BSD-3',
    author="Qiuyu Lian",
    author_email="ql333@cam.ac.uk",
    description="A tool for TOtal-MRNA-Aware Single-cell RNA-seq data analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    #package_dir={'': 'tomas'},
    url="https://tomas.readthedocs.io/en/latest/",
    #python_requires='>=3.8',
    install_requires=[
        "pyDIMM==0.1.0",
#        "numpy",
#        "scipy",
#        "tqdm",
#        "multiprocessing",
        "scanpy>=1.7.1",
#        "pickle",
#        "statsmodels"
    ]
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul  6 03:47:01 2022

@author: qy
"""
import time, os
import numpy as np
from scipy.stats import gaussian_kde, norm, dirichlet,multinomial
import multiprocessing as mp
import pandas as pd
import scipy
import scanpy as sc
import copy
from sklearn.mixture import GaussianMixture

#from auxi import rm_outliers
def rm_outliers(x):
    iqr = scipy.stats.iqr(x)
    outlier_lb = np.quantile(x,0.25)-1.5*iqr
    outlier_ub = np.quantile(x,0.75)+1.5*iqr
    x_shrinkage = x[x > outlier_lb]
    x_shrinkage = x_shrinkage[x_shrinkage<outlier_ub]
    return x_shrinkage#,(outlier_lb,outlier_ub)

#%% Infer Total-mRNA Ratios 


def ratios_bc(adata_mg, dblgroup, max_iter=100, tol=1e-3, n_logR=100, n_p=1000, warm_start=False, verbose=0, verbose_interval=2):
    '''
    Infer the total-mRNA ratio between two cell types given the heterotypic doublets composed by the two cell types.

    Parameters
    ----------
    adata_mg : AnnData
       The UMI count matrix of hetero-doublets in metagenes.
       Rows correspond to droplets and columns to metagenes.
    dblgroup : str
        Name of the heterotypic doublets in the form of 'A_B'.
    max_iter : int, optional
        The maximal number of iterations to perform. The default is 20.
    tol : float, optional
        The convergence threshold. The iterations will stop when the lower bound log-likelihood gain is below this threshold. The default is 1e-3.
    n_logR : int, optional
        The number of R to be sampled per time. The default is 100.
    n_p : int, optional
        The number of p to be sampled per time. The default is 1000.
    warm_start : bool, optional
        If "warm_start" is True, the parameters of the last running is used as initialization for the current call. The default is False.
    verbose : bool, optional
        Enable verbose output. The default is 0, which means no output printed. If 1 then it prints the log-likelihood gain and time lapse for iteration step.
    verbose_interval : int, optional
        Number of iteration done before the next print. The default is 2.

    Returns
    -------
    None. The results are saved in .uns['ratio'] of the input AnnData.

    '''
    
    alpha0,alpha1 = adata_mg.varm['para_diri'].transpose()
    Y = adata_mg.X
    n = len(Y)
    num_trials = Y.sum(1).astype(int)
    
    # Initialization 
    if warm_start:
        
        if 'ratio' not in adata_mg.uns:
            raise ValueError('No fitting history. Run again with "warm_start=False".')

        ratio_dic = adata_mg.uns['ratio']
        log2R_m, log2R_std = ratio_dic['log2R_para']
        
        p0_best = ratio_dic['p_'+dblgroup.split('_')[0]]
        p1_best = ratio_dic['p_'+dblgroup.split('_')[1]]
        w_best = ratio_dic['w_best']
        ll_best = ratio_dic['ll_best']
        log2R_std_prior = ratio_dic['log2R_std_prior']
 
    else:
    
        w_prior = initialize_w(Y, alpha0, alpha1)
        log2R_m = np.log2((1-w_prior)/w_prior)
        #log2R_std_prior = adata_mg.uns['para_logUMI'].loc[dblgroup.split('_'),'std'].sum()
        #log2R_std = copy.deepcopy(log2R_std_prior)
        log2R_std = np.sqrt(np.sum(adata_mg.uns['para_logUMI'].loc[dblgroup.split('_'),'std']**2))
        log2R_std_prior = np.max([adata_mg.uns['para_logUMI'].loc[dblgroup,'std'], log2R_std])
        log2R_std = copy.deepcopy(log2R_std_prior)
        
        p0_best = np.array([alpha0/alpha0.sum()]*n)
        p1_best = np.array([alpha1/alpha1.sum()]*n)
        #w_best = np.array([w_prior]*n)
        logR_sampling = np.random.normal(log2R_m, log2R_std, n)
        w_best = 1/(2**logR_sampling+1)
        
        #p_dbl_best = alpha0/alpha0.sum()*w_prior+alpha1/alpha1.sum()*(1-w_prior)
        p_dbl_best = np.array([alpha0/alpha0.sum()*w + alpha1/alpha1.sum()*(1-w) for w in w_best])
        
        ll_best = np.array([[dirichlet.logpdf(alpha0/alpha0.sum(),alpha0)]*n, 
                            [dirichlet.logpdf(alpha1/alpha1.sum(),alpha1)]*n,
                            list(norm.logpdf(logR_sampling,log2R_m,log2R_std)),
                            [multinomial.logpmf(Y[j],num_trials[j],p_dbl_best[j]) for j in range(n)]])
        
        ratio_dic = {}
        ratio_dic['p_'+dblgroup.split('_')[0]] = p0_best
        ratio_dic['p_'+dblgroup.split('_')[1]] = p1_best
        ratio_dic['w_best'] = w_best
        ratio_dic['log2R_para'] = [log2R_m, log2R_std]
        ratio_dic['R_est'] = 2**log2R_m
        ratio_dic['ll_best'] = ll_best.sum(0)
        ratio_dic['niter'] = 0
        adata_mg.uns['ratio'] = ratio_dic
            
            
    ll_tot_best = ll_best.sum()
    delta_ll = np.inf
    print('Initialized.')
    
    #niter = 0
    niter_NoUpdate = 0
    t0 = time.time()
    ll = copy.deepcopy(ll_tot_best)
    for niter in range(max_iter):

        NoUpdate = 0
        ##### update w, fix p
        # logR_sampling = np.random.normal(log2R_m, log2R_std_prior, n_logR)
        # logR_ll = norm.logpdf(logR_sampling, log2R_m, log2R_std_prior) 
        logR_sampling = np.random.normal(log2R_m, log2R_std, n_logR)
        logR_ll = norm.logpdf(logR_sampling, log2R_m, log2R_std) 
        W = 1/(2**logR_sampling+1)

        p_dbl_fixp = np.array([p0_best*w + p1_best*(1-w) for w in W])

        mult_ll_allsampling = np.array([[multinomial.logpmf(Y[j],num_trials[j],p_dbl_fixp[i,j]) for i in range(n_logR)] for j in range(n)])
        ll_mult_logR = mult_ll_allsampling+logR_ll
        ll_mult_logR_max = ll_mult_logR.max(1)

        midx = ll_mult_logR.argmax(1)
        w_best_tmp = W[midx] 
        mult_ll_max = np.array([mult_ll_allsampling[i][midx[i]] for i in range(n)])
        logR_ll_max = logR_ll[midx]

        replace = ll_mult_logR_max > ll_best[2:].sum(0)
        if sum(replace) > 0:

            w_best[replace] = w_best_tmp[replace]
            log2R_m,log2R_std = norm.fit(rm_outliers(np.log2((1-w_best)/w_best)))
            #log2R_m,log2R_std = norm.fit(np.log2((1-w_best)/w_best))
            #if log2R_std < 0.01: #
            log2R_std = np.max([log2R_std,log2R_std_prior])

            ll_best[2,replace] = logR_ll_max[replace]
            ll_best[3,replace] = mult_ll_max[replace]

            ll_tot = ll_best.sum()
            delta_ll = ll_tot - ll_tot_best
            ll_tot_best = ll_tot
        else:    
            NoUpdate += 0.5

        ##### update p, fix w
        p0 = np.random.dirichlet(alpha0, n_p)
        diri0_ll = np.array([dirichlet.logpdf(p, alpha0) for p in p0])

        p1 = np.random.dirichlet(alpha1, n_p)
        diri1_ll = np.array([dirichlet.logpdf(p, alpha1) for p in p1])

        p_dbl_fixw = np.array([p0*w + p1*(1-w) for w in w_best])

        mult_ll_allsampling = np.array([[multinomial.logpmf(Y[j],num_trials[j],p_dbl_fixw[j,i]) for i in range(n_p)] for j in range(n)])
        ll_diri_mult = mult_ll_allsampling+diri0_ll+diri1_ll
        ll_diri_mult_max = ll_diri_mult.max(1)

        midx = ll_diri_mult.argmax(1)
        p0_best_tmp = p0[midx]
        p1_best_tmp = p1[midx]
        mult_ll_max = np.array([mult_ll_allsampling[i,midx[i]] for i in range(n)])
        diri0_ll_max = diri0_ll[midx]
        diri1_ll_max = diri1_ll[midx]

        replace = ll_diri_mult_max > ll_best[[0,1,3]].sum(0)

        if sum(replace) > 0:

            p0_best[replace] = p0_best_tmp[replace]
            p1_best[replace] = p1_best_tmp[replace]

            ll_best[0,replace] = diri0_ll_max[replace]
            ll_best[1,replace] = diri1_ll_max[replace]
            ll_best[3,replace] = mult_ll_max[replace]

            ll_tot = ll_best.sum()
            delta_ll = ll_tot - ll_tot_best
        else:
            NoUpdate += 0.5

        if NoUpdate == 1:
            niter_NoUpdate += 1
        else:
            niter_NoUpdate = 0
         
        if verbose > 0 and (niter+1) % verbose_interval == 0:
            print('Iteration '+str(niter+1)+'\ttime lapse '+str(time.time()-t0)+'s\tll change '+str(ll_best.sum()-ll)+'\n')
            t0 = time.time()
            ll = ll_best.sum()
        if niter_NoUpdate >= 3 or delta_ll < tol:
            break
    
    # update the info stored in anndata
    adata_mg.uns['ratio']['p_'+dblgroup.split('_')[0]] = p0_best
    adata_mg.uns['ratio']['p_'+dblgroup.split('_')[1]] = p1_best
    adata_mg.uns['ratio']['w_best'] = w_best
    adata_mg.uns['ratio']['log2R_para'] = [log2R_m, log2R_std]
    adata_mg.uns['ratio']['R_est'] = 2**log2R_m
    adata_mg.uns['ratio']['ll_best'] = ll_best
    adata_mg.uns['ratio']['niter'] += niter

    ## check confidence
    r_list = (1-w_best)/w_best
    if len(r_list) < 15:
        log2r = np.log2(r_list)
    else:
        log2r = rm_outliers(np.log2(r_list))

    unimodal = GaussianMixture(1).fit(log2r.reshape(-1,1))
    bimodal =  GaussianMixture(2).fit(log2r.reshape(-1,1))
    lrt = -2*(unimodal.score_samples(log2r.reshape(-1,1)).sum() - bimodal.score_samples(log2r.reshape(-1,1)).sum())
    pval = scipy.stats.chi2.sf(lrt, 3)
    adata_mg.uns['ratio']['confidence'] = pval



    
from scipy import optimize

def func(x, a):
    y = a*x 
    return y


def initialize_w(Y, alpha1, alpha2, method='kde'):
    
    p1 = alpha1/sum(alpha1)
    p2 = alpha2/sum(alpha2)
    gsum_mix = Y.sum(0)
    pmix = gsum_mix/sum(gsum_mix)
    if method=='ols':
        n = Y.shape[0]
        w_ = optimize.curve_fit(func, 
                                xdata = np.ravel(np.array([p1]*n)-np.array([p2]*n)), 
                                ydata = np.ravel(pmix - np.array([p2]*n)))
        w = w_[0][0]
        if w >= 1 or w <= 0:
            method=='kde'
    
    if method=='kde':
        idx_in = [i for i in range(len(p1)) if pmix[i] > min(p1[i],p2[i]) and pmix[i] < max(p1[i], p2[i])]
        
        w_tmp = [(pmix[i]-p2[i])/(p1[i]-p2[i]) for i in idx_in if p1[i] - p2[i] != 0]
        
        kde = gaussian_kde(w_tmp)
        samples = np.linspace(0, 1, 1000)
        probs = kde.evaluate(samples)
        maxima_index = probs.argmax()
        w = samples[maxima_index]

    return w




#%% Extract Exclusive Meta-Genes  

def get_dbl_mg(adata,groupby,groups='all',output=None, num_mg = 100, kl_cutoff=None, merging_threshold=5, skip_threshold=2,alphaMin = 1):
    '''
    Extract metagenes from raw UMI counts of heterotypic doublets.

    Parameters
    ----------
    adata : AnnData
        The (annotated) UMI count matrix of shape `n_obs` × `n_vars`.
        Rows correspond to droplets and columns to genes.
    groupby : str
        The key of the droplet categories stored in adata.obs. 
    groups : list or str ('all'), optional
        Groups of heterodbls to extract metagenes. List of a groups of valid dbl names or a string 'all'. The default is 'all'.
    output : path
        Path to save the results.
    num_mg : int, optional
        Number of exclusive meta-genes. The default is 100.
    kl_cutoff : float, optional
        DESCRIPTION. The default is 1.
    merging_threshold : float, optional
        Stop merging when alpha sum of metagenes exceeds the threshold in conterpart celltype. The default is 5.
    skip_threshold : float, optional
         If adding a gene into current metegene leads to 'skip_threshold'-fold change of the std of alpha values,
         this gene is skipped. The default is 2.
    alphaMin : float, optional
         Individual genes with alpha greater than the threshold are skipped in mergging step and considerted as a specifial metagene with one member gene. The default is 1.

    Raises
    ------
    ValueError
        Check the legality of argements.

    Returns
    -------
    adata_mgdic : dic
        Keys are names of heterotypic doublets. Values are UMI counts in metagenes of heterotypic doublets.

    '''
    if isinstance(groups, str):
        if groups == 'all':
            #groups = adata.obs[groupby].unique().tolist()
            groups = [d for d in adata.obs[groupby].unique() if len(d.split('_'))==2]
            if 'unknown' in groups:
                groups.remove('unknown')
        else:
            raise ValueError('Invalid input. If "groups" is a string, it can only be "all". If you want to specify a single heterodbl group, put it in a list, e.g. ["B_CD8T"].')
    elif isinstance(groups, list):
        for f in groups:
            if f not in adata.obs[groupby].unique():
                raise ValueError('Invalid input. The annotation of heterodbls, set by argument "groupby" in adata.obs, do not contain "'+f+'".')
    else:
        raise ValueError('Invalid input of argument "groupby".')
        
    if isinstance(kl_cutoff,float):
        if 'kl' not in adata.varm.keys():
            raise ValueError('KL divergences have not be calculated yet. Run "get_dbl_mg" with "kl_cutoff" set as "None" to disabe filtering genes by KL (recommended) or first calculate the KL divergens')

    adata_mgdic = {}
    for dblgroup in groups:
        mg_tmp = get_dbl_mg_bc(adata, 
                               groupby, 
                               dblgroup, 
                               output=output,
                               num_mg=num_mg, 
                               kl_cutoff=kl_cutoff,
                               merging_threshold=merging_threshold,
                               skip_threshold=skip_threshold,
                               alphaMin=alphaMin)
        adata_mgdic[dblgroup] = mg_tmp

    return adata_mgdic
    





def get_dbl_mg_bc(adata, groupby, dblgroup, output=None, num_mg = 100, kl_cutoff=None, merging_threshold=5, skip_threshold=2,alphaMin = 1):
    '''
    Extract metagenes from raw UMI counts of a certain type of heterotypic doublets.
    Parameters
    ----------
    adata : AnnData
        The (annotated) UMI count matrix of shape `n_obs` × `n_vars`.
        Rows correspond to droplets and columns to genes.
    groupby : str
        The key of the droplet categories stored in adata.obs. 
    groups : list of strings
        Specify two droplet categories, e.g. ['Homo-ct1', 'Homo-ct2'] annotated in adata.obs[groupby], which form the hetero-dbl of interest.
    output : path
        Path to save the results.
    num_mg : int, optional
        Number of exclusive meta-genes. The default is 100.
    kl_cutoff : float, optional
        DESCRIPTION. The default is 1.
    merging_threshold : float, optional
        Stop merging when alpha sum of metagenes exceeds the threshold in conterpart celltype. The default is 5.
    skip_threshold : float, optional
         If adding a gene into current metegene leads to 'skip_threshold'-fold change of the std of alpha values,
         this gene is skipped. The default is 2.
    alphaMin : float, optional
         Individual genes with alpha greater than the threshold are skipped in mergging step and considerted as a specifial metagene with one member gene. The default is 1.
    Returns
    -------
    adata_dbl_mg_top : AnnData
        The UMI count matrix of shape `n_obs` × `n_metangenes`.
        Rows correspond to droplets and columns to metagenes.
    '''
    groups = dblgroup.split('_') #+[dblgroup]

    adata_dbl = adata[adata.obs[groupby]==dblgroup,:].copy()
    adata_dbl = adata_dbl[:,adata_dbl.X.sum(0)>0]


    #count_dbl = adata_dbl[adata_dbl.obs[groupby]==dblgroup,:].X
    count_dbl = adata_dbl.X
    if isinstance(count_dbl, scipy.sparse.csr_matrix):
        count_dbl = count_dbl.toarray()


    raw_Alpha1_0 = adata_dbl.varm['para_diri'][groups[0]]
    raw_Alpha2_0 = adata_dbl.varm['para_diri'][groups[1]]


    if isinstance(kl_cutoff,float):
        # use genes with KL greater than cutoff to obtain meta-genes
        kl = adata_dbl.varm['kl']
        KL_filter_idx = np.where(kl > kl_cutoff)[0]
    else:
        KL_filter_idx = np.array([i for i in range(len(raw_Alpha1_0))])

    raw_Alpha1 = raw_Alpha1_0[KL_filter_idx]
    raw_Alpha2 = raw_Alpha2_0[KL_filter_idx]

    mg_dic = get_mg(raw_Alpha1, raw_Alpha2, merging_threshold, skip_threshold,alphaMin)

    mg_pool = mg_dic['mg_pool'] 

    mg_gidx = [list(KL_filter_idx[term]) for term in mg_pool]
    mg_genepool = [g for sub in  mg_gidx  for g in sub]

    mg_alpha1 = mg_dic['alpha1']
    mg_alpha2 = mg_dic['alpha2']

    left_gidx = list(set(range(len(raw_Alpha1_0))).difference(mg_genepool)) 
    mg_gidx.append(left_gidx)

    # merge alpha and counts according to meta-genes

    mg_Alpha_arr = np.concatenate((np.array([mg_alpha1, mg_alpha2]), np.array([[sum(raw_Alpha1_0)-sum(mg_alpha1)],[sum(raw_Alpha2_0)-sum(mg_alpha2)]])),axis=1)
    Y = np.transpose(np.array([np.sum(count_dbl[:,term],1) for term in mg_gidx]))

    adata_dbl_mg = sc.AnnData(Y,dtype=Y.dtype)
    adata_dbl_mg.obs_names = adata_dbl.obs_names#[adata_dbl.obs[groupby]==dblgroup]
    adata_dbl_mg.varm['para_diri'] = mg_Alpha_arr.transpose()
    if output is not None:
        if not os.path.exists(output):
            os.makedirs(output)
        adata_dbl_mg.write_h5ad(os.path.join(output,dblgroup+'_dbl_mg.h5ad'))

    # extract top num_mg 
    topmg_idx = [[i] for i in range(num_mg)] + [list(range(num_mg,  len(mg_gidx)))]
    mg_Alpha_new = np.transpose(np.array([np.sum(mg_Alpha_arr[:,term],1) for term in topmg_idx]))
    Y_new = np.transpose(np.array([np.sum(Y[:,term],1) for term in topmg_idx]))

    adata_dbl_mg_top = sc.AnnData(Y_new, dtype=Y_new.dtype)
    adata_dbl_mg_top.obs_names = adata_dbl.obs_names#[adata.obs[groupby]==dblgroup]
    adata_dbl_mg_top.varm['para_diri'] = mg_Alpha_new.transpose()
    if len(adata.uns.keys()):
        for val in adata.uns.keys():
            adata_dbl_mg_top.uns[val] = adata.uns[val]

    return adata_dbl_mg_top





get_cosine = lambda a: sum(a)/(np.linalg.norm(a)*np.sqrt(2))


def get_mg(raw_Alpha1, raw_Alpha2, merging_threshold=1, skip_threshold=2, alphaMin = 1):
    """
    fuse genes with alpha < 1 into metagenes in a balanced way

    Parameters
    ----------
    raw_Alpha1 : arr
        DESCRIPTION.
    raw_Alpha2 : arr
        DESCRIPTION.
    merging_threshold : float, optional
        DESCRIPTION. The default is 1.
    skip_threshold : float, optional
        DESCRIPTION. The default is 2.

    Returns
    -------
    mg_dic : TYPE
        DESCRIPTION.

    """
    ## step 1: merge dims with tiny alpha to get metagenes (alpha > 1)
    
    raw_Alpha_arr = np.array([raw_Alpha1, raw_Alpha2])
    raw_cosine_vals = np.array([get_cosine(g) for g in raw_Alpha_arr.transpose()])
    
    # rank genes acording to the angle to (1,1)
    order_by_cosine = np.argsort(raw_cosine_vals)
    
    raw_set_1 = [g for g in order_by_cosine if raw_Alpha_arr[0,g] > raw_Alpha_arr[1,g]] 
    raw_set_2 = [g for g in order_by_cosine if raw_Alpha_arr[0,g] < raw_Alpha_arr[1,g]] 
    
    # extract dimensions with alpha > 1
    mg_single_set_1 = [g for g in raw_set_1 if raw_Alpha_arr[0,g] > alphaMin and raw_Alpha_arr[1,g] > alphaMin] 
    mg_single_set_2 = [g for g in raw_set_2 if raw_Alpha_arr[1,g] > alphaMin and raw_Alpha_arr[0,g] > alphaMin] 
    
    MetaGene_1_single = [[g] for g in mg_single_set_1]
    MetaGene_2_single = [[g] for g in mg_single_set_2]


    # set dimensions with alpha < 1 as seeds for merging meta genes
    mg_1_seeds = [g for g in raw_set_1 if g not in mg_single_set_1] 
    mg_2_seeds = [g for g in raw_set_2 if g not in mg_single_set_2] 
    
    # merge genes with smaller alpha in 1
    mg_1_merge = []
    while len(mg_1_seeds):
        
        # mg_1_seeds mean genes have larger alpah in 1 than 2
        # so merge genes in 1 until their alpha sum is greater than threshold in 2
        conterpart =  np.cumsum(raw_Alpha_arr[1, mg_1_seeds])
        
        idx = np.where(conterpart > merging_threshold)
        # idx = np.where(cumsum_1 > 1)
        if len(idx[0]) > 0:
            idx = idx[0][0]
            mg_1_merge.append(mg_1_seeds[:idx+1])
            del mg_1_seeds[:idx+1]
            
        else:
            # if left genes whose alpha sum < 1, add them into the last metagene
            if len(mg_1_merge) > 0:
                term = mg_1_merge.pop()
                term = term + mg_1_seeds
                mg_1_merge.append(term)
                # mg_1_seeds = []
            else:
                # if sum of all seed genes < merging_threshold, merge them with minimal single genes
                s_1_minidx = raw_Alpha1[mg_single_set_1].argmin()
                mg_1_merge = mg_1_seeds + MetaGene_1_single[s_1_minidx]
            
                del MetaGene_1_single[s_1_minidx], mg_single_set_1[s_1_minidx]
            
            break
        
    # merge genes with smaller alpha in 2
    mg_2_merge = []
    while len(mg_2_seeds):
        
        conterpart =  np.cumsum(raw_Alpha_arr[0, mg_2_seeds])
        
        idx = np.where(conterpart > merging_threshold)
        # idx = np.where(cumsum_2 > 1)
        if len(idx[0]) > 0:
            idx = idx[0][0]
            mg_2_merge.append(mg_2_seeds[:idx+1])
            del mg_2_seeds[:idx+1]
            
        else:
            # if left genes whose alpha sum < 1, add them into the last metagene
            if len(mg_2_merge) > 0:
                term = mg_2_merge.pop()
                term = term + mg_2_seeds
                mg_2_merge.append(term)
            else:
                # if sum of all seed genes < merging_threshold, merge them with minimal single genes
                s_2_minidx = raw_Alpha2[mg_single_set_2].argmin()
                mg_2_merge = mg_2_seeds + MetaGene_2_single[s_2_minidx]
            
                del MetaGene_2_single[s_2_minidx], mg_single_set_2[s_2_minidx]

            break
    
    
    if len(mg_1_merge) > 0:
        if isinstance(mg_1_merge[0],list):
            MetaGene_1 = mg_1_merge + MetaGene_1_single
        else:
            ### all genes with small alpha are merged int one
            MetaGene_1 = [mg_1_merge] + MetaGene_1_single
    else:
        MetaGene_1 = MetaGene_1_single
    
    
    if len(mg_2_merge) > 0:
        if isinstance(mg_2_merge[0],list):
            MetaGene_2 = mg_2_merge + MetaGene_2_single
        else:
            ### all genes with small alpha are merged int one
            MetaGene_2 = [mg_2_merge] + MetaGene_2_single        
    else:
        MetaGene_2 = MetaGene_2_single
    
    
    
    # metagenes, list of lists, recording the index of genes composing each meta gene 
    mg_list = MetaGene_1 + MetaGene_2
    
    
    ## step 2: rank metagenes in probabilistically balanced way

    mg_Alpha_arr = np.array([[sum(raw_Alpha_arr[0,term]) for term in mg_list], 
                              [sum(raw_Alpha_arr[1,term]) for term in mg_list]])

    mg_cosine2XequalY = np.array([get_cosine(g) for g in mg_Alpha_arr.transpose()])
    
    mg_order_by_cosine = list( np.argsort(mg_cosine2XequalY) )
    
    mg_order_by_cosine.reverse()
    
    mg_idx_XgreaterY = [g for g in mg_order_by_cosine if mg_Alpha_arr[0,g] > mg_Alpha_arr[1,g]]
    mg_idx_YgreaterX = [g for g in mg_order_by_cosine if mg_Alpha_arr[0,g] <= mg_Alpha_arr[1,g]]

    
    mg_idx = [mg_idx_XgreaterY.pop(), mg_idx_YgreaterX.pop()]
    
    mg_Alpha_reordered = mg_Alpha_arr[:,mg_idx]
    
    cur_sum_x, cur_sum_y = mg_Alpha_reordered.sum(1)
    cur_std_x, cur_std_y = mg_Alpha_reordered.std(1)
    
    skip_idx_x = []
    skip_idx_y = []
    
    while len(mg_idx_XgreaterY)*len(mg_idx_YgreaterX):
        
        while cur_sum_x > cur_sum_y and len(mg_idx_YgreaterX) > 0:
            
            cur_mg_idx_y = mg_idx_YgreaterX.pop()
            cur_std_y_new = np.std(mg_Alpha_arr[1,mg_idx+[cur_mg_idx_y]])
            
            if cur_std_y_new/cur_std_y < skip_threshold:
            
                mg_idx.append(cur_mg_idx_y)
                mg_Alpha_reordered = mg_Alpha_arr[:,mg_idx]
                cur_sum_x, cur_sum_y = mg_Alpha_reordered.sum(1)
                cur_std_y = cur_std_y_new
            else:
                skip_idx_y.append(cur_mg_idx_y)
  
        while cur_sum_y > cur_sum_x and len(mg_idx_XgreaterY) > 0:
            
            cur_mg_idx_x = mg_idx_XgreaterY.pop()
            cur_std_x_new = np.std(mg_Alpha_arr[0,mg_idx+[cur_mg_idx_x]])
            
            if cur_std_x_new/cur_std_x < skip_threshold:
            
                mg_idx.append(cur_mg_idx_x)
                mg_Alpha_reordered = mg_Alpha_arr[:,mg_idx]
                cur_sum_x, cur_sum_y = mg_Alpha_reordered.sum(1)
                cur_std_x = cur_std_x_new
            else:
                skip_idx_y.append(cur_mg_idx_x)
  

    balanced_mg_num = len(mg_idx)
    
    mg_idx_all = mg_idx + skip_idx_y+ mg_idx_YgreaterX + skip_idx_x + mg_idx_XgreaterY
    mg_Alpha_reordered = mg_Alpha_arr[:,mg_idx_all]

    # list of lists, recording the index of genes composing each meta gene after reordering
    reordered_mg_pool = [mg_list[g] for g in mg_idx_all]
    #reordered_gidx_pool = [g for sub in reordered_mg_pool for g in sub]
    
    #mg_left = [mg_list[g] for g in range(len(mg_list)) if g not in mg_idx]
    #gidx_left = [g for sub in mg_left for g in sub]
    
    mg_dic = {'alpha1':mg_Alpha_reordered[0],
              'alpha2':mg_Alpha_reordered[1],
              'mg_pool':reordered_mg_pool,
              #'gidx_pool':reordered_gidx_pool,
              'balanced_mg_num':balanced_mg_num}
              #'mg_left':mg_left,
              #'gidx_left':gidx_left}
    
    return mg_dic





#%% preparation in scRNA-only scenario 


from matplotlib import pyplot as plt
import seaborn as sns
import itertools
import matplotlib.image as mpimg
sns.set_style("ticks")


def heteroDbl(adata, d_groupby, ct_groupby, de_sorted=None,dbl_groupby = 'heteroDbl',vis=False):
    '''
    Identification and refinement of heterotypic doublets

    Parameters
    ----------
    adata : AnnData
        UMI matrix of all droplet population.
    d_groupby : str
        The key of the droplet annotation ('Homotypic' or 'Heterotypic') results stored in adata.obs.
    ct_groupby : str
        The key of the cell type annotation results stored in adata.obs.
    de_sorted : data frame, optional
        Output of function 'extract_specific_genes'. The default is None.
    dbl_groupby : TYPE, optional
        DESCRIPTION. The default is 'heteroDbl'.
    vis : bool, optional
        Whether to visualize the heterodlb identification results. The default is True.

    Returns
    -------
    None. Identified heterotypic 

    '''
    
    groups = adata.obs[ct_groupby].unique().tolist()

    if 'unknown' in groups:
        groups.remove('unknown')

    #if dbl_groupby not in adata.obs:
    adata.obs[dbl_groupby] = 'unknown'
        
    for pair in itertools.combinations(groups,2):
        dbl = '_'.join(pair)
        print('Detect doublets composed by cell type '+pair[0]+' and '+pair[1]+'.')
        dbl_candidate = heteroDbl_bc(adata, dbl,  d_groupby = d_groupby, ct_groupby = ct_groupby, de_sorted = de_sorted,vis=vis)
        adata.obs.loc[dbl_candidate,dbl_groupby] = dbl  
        print(str(len(dbl_candidate))+' putative heterotypic doublets detected.')        




def heteroDbl_bc(adata, dbl, d_groupby, ct_groupby, g2meta = None, de_sorted=None, mg_gnum=[10,10], threshold = None, threshold_x=None,threshold_y=None,vis=True, log=True, return_fig=False, dpi=80):
    '''
    Identification and refinement of a certain type of heterotypic doublets specified by "dbl".

    Parameters
    ----------
    adata : AnnData
        UMI matrix of all droplet population.
    dbl : str
        Heterotypic dboulet group to be investigated, named by two cell types linked by '_'.
    d_groupby : str
        The key of the droplet annotation ('Homotypic' or 'Heterotypic') results stored in adata.obs.
    ct_groupby : str
        The key of the cell type annotation results stored in adata.obs.
    g2meta : list, optional
        A list of genes. The default is None.
    de_sorted : data frame, optional
        Output of function 'extract_specific_genes'. The default is None.
    mg_gnum : list, optional
        Number of DEG merged to identify heterodbls. The default is [10,10].
    threshold : list, optional
        Cutoff of expression levels in current cell type pair to identify heterodbls. The default is None.
    threshold_x : float, optional
        Cutoff of expression level in current (x axis) cell type to identify heterodbls. The default is None.
    threshold_y : float, optional
        Cutoff of expression level in current (y axis) cell type to identify heterodbls. The default is None.
    vis : bool, optional
        Whether to visualize the heterodlb identification results. The default is True.
    log : bool, optional
        Whether to calculate in logarithmic form. The default is True (recommended).
    return_fig : bool, optional
        Whether to return the figure object. The default is False.

    Raises
    ------
    ValueError
        Inspect if genes to be merged for heterotypic doublet detection are specified in a proper way.

    Returns
    -------
    list
        Idnetified heterotypic doublets composed by input cell type pair.

    '''
    
    if 'dblPred_'+dbl in adata.obs:
        del adata.obs['dblPred_'+dbl]
    
    ct_pair = dbl.split('_')
    if g2meta is not None:
        g2meta_x,g2meta_y = g2meta
    elif de_sorted is not None:
        g2meta_x = de_sorted[ct_pair[0]+'_n'][:mg_gnum[0]]
        g2meta_y = de_sorted[ct_pair[1]+'_n'][:mg_gnum[0]]
    else:
        raise ValueError("value missing. please assign genes to merge by para 'g2meta' or 'de_sorted'")
    
    adata_sgl = adata[adata.obs[d_groupby]=='Homotypic'].copy()
    adata_dbl = adata[adata.obs[d_groupby]=='Heterotypic'].copy()
    
    x_dbl = adata_dbl[:,g2meta_x].X.toarray().sum(1)
    x_sgl = adata_sgl[:,g2meta_x].X.toarray().sum(1)

    y_dbl = adata_dbl[:,g2meta_y].X.toarray().sum(1)
    y_sgl = adata_sgl[:,g2meta_y].X.toarray().sum(1)

    if log:
        x_dbl = np.log(x_dbl+1)
        x_sgl = np.log(x_sgl+1)
        y_sgl = np.log(y_sgl+1)
        y_dbl = np.log(y_dbl+1)

    dbl_sub = pd.DataFrame({'x':x_dbl,'y':y_dbl},index=adata_dbl.obs_names)

    sgl_sub = pd.DataFrame({'x':x_sgl,'y':y_sgl,
                            'celltype_pair':['others']*len(x_sgl)
                           },
                           index=adata_sgl.obs_names)
    sgl_sub.loc[adata_sgl.obs[ct_groupby]==ct_pair[0],'celltype_pair'] = ct_pair[0]
    sgl_sub.loc[adata_sgl.obs[ct_groupby]==ct_pair[1],'celltype_pair'] = ct_pair[1]
    sgl_sub['celltype_pair'] = pd.Categorical(sgl_sub['celltype_pair'].values.tolist(),categories=ct_pair+['others'])
    sgl_sub[ct_groupby] = adata_sgl.obs[ct_groupby]

    if threshold is None:
        ct_x_mgy = sgl_sub.loc[sgl_sub['celltype_pair']==ct_pair[0], 'y']
        ct_x_mgx = sgl_sub.loc[sgl_sub['celltype_pair']==ct_pair[0], 'x']

        ct_y_mgx = sgl_sub.loc[sgl_sub['celltype_pair']==ct_pair[1], 'x']
        ct_y_mgy = sgl_sub.loc[sgl_sub['celltype_pair']==ct_pair[1], 'y']

        mgy_threshold = np.max(ct_x_mgy[ct_x_mgx > np.max(ct_y_mgx)])
        mgx_threshold = np.max(ct_y_mgx[ct_y_mgy > np.max(ct_x_mgy)])
        
        mgx_threshold += np.max([0.1*mgx_threshold, 0.2]) # add a margin
        mgy_threshold += np.max([0.1*mgy_threshold, 0.2]) # add a margin
        
    else:
        mgx_threshold, mgy_threshold = threshold
        
    if threshold_x is not None:
        mgx_threshold = threshold_x
        
    if threshold_y is not None:
        mgy_threshold = threshold_y
    

    dbl_onhit = [d for d in dbl_sub.index if dbl_sub.loc[d,'x'] > mgx_threshold and dbl_sub.loc[d,'y'] > mgy_threshold]
    adata.obs['dblPred_'+dbl] = 0
    adata.obs.loc[dbl_onhit,'dblPred_'+dbl] = 1
    
    if vis:
        
        dbl_sub['pred'] = ['others']*dbl_sub.shape[0]
        dbl_sub.loc[dbl_onhit,'pred'] = '_'.join(ct_pair)
        dbl_sub['pred'] = pd.Categorical(dbl_sub['pred'],categories=['_'.join(ct_pair), 'others'])


        grid1 = sns.JointGrid(data=sgl_sub, x="x", y="y")
        g1 = grid1.plot_joint(sns.scatterplot, hue='celltype_pair', data=sgl_sub, palette=['blue','green','gray'])
        sns.kdeplot(x=sgl_sub['x'], ax=g1.ax_marg_x, legend=False)
        sns.kdeplot(y=sgl_sub['y'], ax=g1.ax_marg_y, legend=False)
        g1.ax_joint.axhline(y=mgy_threshold,c='blue')
        g1.ax_joint.axvline(x=mgx_threshold,c='green')
        g1.set_axis_labels('MetaDEG of A (putative singlets)', 'MetaDEG of B (putative singlets)', fontsize=18)

        grid2 = sns.JointGrid(data=dbl_sub, x="x", y="y")
        g2 = grid2.plot_joint(sns.scatterplot, hue='pred', data=dbl_sub, palette=['red','gray'])
        sns.kdeplot(x=dbl_sub['x'], ax=g2.ax_marg_x, legend=False)
        sns.kdeplot(y=dbl_sub['y'], ax=g2.ax_marg_y, legend=False)
        g2.ax_joint.axhline(y=mgy_threshold,c='blue')
        g2.ax_joint.axvline(x=mgx_threshold,c='green')
        g2.set_axis_labels('MetaDEG of A (putative doublets)', 'MetaDEG of B (putative doublets)', fontsize=18)

        ############### save plots in memory temporally
        g1.savefig('g1.png')
        plt.close(g1.fig)

        g2.savefig('g2.png')
        plt.close(g2.fig)

        ############### create subplots from temporay memory
        fig, axarr = plt.subplots(1, 2, figsize=(10, 5),dpi=dpi)
        axarr[0].imshow(mpimg.imread('g1.png'))
        axarr[1].imshow(mpimg.imread('g2.png'))

        # turn off x and y axis
        [ax.set_axis_off() for ax in axarr.ravel()]

        plt.tight_layout()
        plt.show()
        
        if return_fig:
            #return dbl_onhit, fig
            return fig






def heterodbl_summary(adata,ct_groupby):
    
    dblpred_df = adata.obs[[c for c in adata.obs.columns if c.startswith('dblPred_')]]
    dbl_list = dblpred_df.index[dblpred_df.sum(1) == 1]
    adata.obs.loc[dbl_list, ct_groupby] = [dblpred_df.columns[dblpred_df.loc[d,:].argmax()].replace('dblPred_', '') for d in dbl_list]

    if 'log2_total_UMIs' not in adata.obs:
        adata.obs['log2_total_UMIs'] = np.log2(np.ravel(adata.X.sum(1)))   

    sgl_groups = [v for v in adata.obs[ct_groupby].unique() if len(v.split('_'))==1 and v!='unknown']
    dbl_groups = [v for v in adata.obs[ct_groupby].unique() if len(v.split('_'))==2]

    sgl_m = pd.Series([adata.obs.loc[adata.obs[ct_groupby]==c, 'log2_total_UMIs'].mean() for c in sgl_groups], index = sgl_groups)
    dbl_m = pd.Series([adata.obs.loc[adata.obs[ct_groupby]==c, 'log2_total_UMIs'].mean() for c in dbl_groups], index = dbl_groups)

    sgl_vote = pd.DataFrame(np.zeros([len(sgl_groups),len(sgl_groups)]),index=sgl_groups,columns=sgl_groups)
    dbl_vote = pd.DataFrame(np.zeros([len(sgl_groups),len(sgl_groups)]),index=sgl_groups,columns=sgl_groups)

    for pair in itertools.combinations(sgl_groups, 2):

        c1,c2 = pair
        sgl_vote.loc[c1,c2] = sgl_m[c1] - sgl_m[c2]
        sgl_vote.loc[c2,c1] = sgl_m[c2] - sgl_m[c1]


    for pair in itertools.combinations(dbl_m.index, 2):

        p1, p2 = pair
        #print(p1,p2)
        p1_cts = p1.split('_')
        p2_cts = p2.split('_')

        vals,cnts = np.unique(p1_cts+p2_cts,return_counts=True)
        if np.any(cnts>1):
            p1_cts.remove(vals[cnts>1])
            p2_cts.remove(vals[cnts>1])
            #print(dbl_m[p1] - dbl_m[p2])
            dbl_vote.loc[p1_cts[0],p2_cts[0]] = dbl_m[p1] - dbl_m[p2]
            dbl_vote.loc[p2_cts[0],p1_cts[0]] = dbl_m[p2] - dbl_m[p1] 


    sum_vote = sgl_vote + dbl_vote

    dbl_groups = []
    for pair in itertools.combinations(sgl_groups, 2):
        c1,c2 = pair
        if sum_vote.loc[c1,c2]>0:
            dbl_groups.append(c2+'_'+c1)
        else:
            dbl_groups.append(c1+'_'+c2)

    '''
    for i,v in enumerate(adata.obs[ct_groupby]):
        if v not in dbl_groups+sgl_groups and v!='unknown':
            adata.obs[ct_groupby].iloc[i] = '_'.join([v.split('_')[1],v.split('_')[0]])
    '''
    
    for d in adata.obs_names:
        v = adata.obs.loc[d,ct_groupby]
        if v not in dbl_groups+sgl_groups and v!='unknown':
            adata.obs.loc[d,ct_groupby] = '_'.join([v.split('_')[1],v.split('_')[0]])
            
    vals,cnts=np.unique(adata.obs[ct_groupby], return_counts=True)
    df = pd.DataFrame({'Hetero-dbl-type':vals,'counts':cnts},index=vals)
    print('Counts of each kind of droplets:')
    print(df)





import statsmodels.api as sm

def save_ratio_in_anndata(adata_mgdic, adata, weight=True):
    
    adata.uns['ratio'] = {}
    for dbl in adata_mgdic:
        adata.uns['ratio'][dbl] = adata_mgdic[dbl].uns['ratio']
    
    r_df,r_serial = harmonize_ratios(adata_mgdic,weight=weight)
    adata.uns['ratio_summary'] = r_df
    adata.uns['ratio_serial'] = r_serial
    
    
def harmonize_ratios(adata_mgdic,weight=True):

    dbl_groups = list(adata_mgdic.keys())
    sgl_groups = list(np.unique([d for term in dbl_groups for d in term.split('_')]))

    r_info = pd.DataFrame(np.zeros([len(dbl_groups), len(sgl_groups)]),columns=sgl_groups)

    #print('Ratios estimated from each kind of heterodbls:')
    y_list = []
    for idx,dbl in enumerate(dbl_groups):
        c1,c2 = dbl.split('_')
        r_info.loc[idx, c1] = -1
        r_info.loc[idx, c2] = 1
        r_tmp = adata_mgdic[dbl].uns['ratio']['R_est']
        y_list.append(np.log2(r_tmp))
        #print(c2,':',c1,'=',r_tmp)

    X = r_info.values
    y = np.array(y_list)
    
    if weight:
        mod_wls = sm.WLS(y, X, weights=np.array([adata_mgdic[a].n_obs for a in adata_mgdic]))
    else:
        mod_wls = sm.WLS(y, X, weights=np.array([1]*len(adata_mgdic)))
    res_wls = mod_wls.fit()
    # print(res_wls.summary())

    pred_ols = res_wls.get_prediction()
    # pred_ols.summary_frame()
    
    #print('\nAfter regression:')
    r_corrected = 2**pred_ols.summary_frame()['mean']
    for idx,dbl in enumerate(dbl_groups):
        c1,c2 = dbl.split('_')
        #print(c2,':',c1,'=',r_corrected[idx])

    r_raw = []
    for dbl in dbl_groups:
        logUMI_para = adata_mgdic[dbl].uns['para_logUMI']
        c1,c2 = dbl.split('_')
        r = 2**(logUMI_para.loc[c2,'mean'] - logUMI_para.loc[c1,'mean'])
        r_raw.append(r)

    r_df = pd.DataFrame({'Ratio_rawUMI':r_raw,
                         'Ratio_estimated':[adata_mgdic[d].uns['ratio']['R_est'] for d in dbl_groups],
                         'Ratio_harmonized':r_adj.values
                        },index=[dbl.split('_')[1]+':'+dbl.split('_')[0] for dbl in dbl_groups])
    
    r_mat = pd.DataFrame(np.ones([len(sgl_groups),len(sgl_groups)]),index=sgl_groups,columns=sgl_groups)
    for pair in r_df.index:
        c1,c2 = pair.split(':')
        r_mat.loc[c1,c2] = r_df.loc[pair,'Ratio_harmonized']
        r_mat.loc[c2,c1] = 1/r_df.loc[pair,'Ratio_harmonized']

    sgl_reference = r_mat.columns[r_mat.apply(lambda col: sum(col>=1)==len(col),axis=0)]
    r_serial = pd.DataFrame(r_mat.loc[:,sgl_reference[0]].sort_values())
    
    return r_df,r_serial





#%% old version of inferring ratios

def ratio_2types(adata,groupby,groups,output,subset=100,nrepeat=10):
    '''
    

    Parameters
    ----------
    adata_dbl_mg : AnnData
        The UMI count matrix of hetero-doublets in metagenes.
        Rows correspond to droplets and columns to metagenes.
    output : path
        Path to save the results.
    nrepeat : int, optional
        Times to repeat synthesizing doublets to infer ratios. The default is 10.

    Returns
    -------
    esti_r_list : numpy.ndarray
        Estimate of total mRNA ratio of each hetero-dbouelt.

    '''
    if not os.path.exists(output):
        os.makedirs(output)
        
    name=groups[2]
    logpath = os.path.join(output,'_'.join(['rlog',name]))
    if not os.path.exists(logpath):
        os.makedirs(logpath)
        
    
    adata_dbl_mg = get_dbl_mg_bc(adata,groupby,groups)
    
    mg_Alpha_new = adata_dbl_mg.varm['para_diri'].transpose()
    Y_new = adata_dbl_mg.X
    
    if subset is not None:
        if Y_new.shape[0] > subset:
            idx = np.random.choice(Y_new.shape[0], subset, replace=False)
            Y_new = Y_new[idx,:]
    
    '''
    ### use raw data to init
    count_dbl = adata[adata.obs[groupby]==groups[2],:].X
    if isinstance(count_dbl, scipy.sparse.csr.csr_matrix):
        count_dbl = count_dbl.toarray()
         
    w_init = initialize_w(count_dbl, adata.varm['para_diri'][groups[0]].values, adata.varm['para_diri'][groups[1]].values)
    ###
    '''
    #w_init = initialize_w(Y_new, mg_Alpha_new[0], mg_Alpha_new[1])
    #estimateW(mg_Alpha_new,Y_new,logpath,w_init,nrepeat,name)
    estimateW(mg_Alpha_new,Y_new,logpath,nrepeat=nrepeat,name=name)

    
    ll_list = []
    for ridx in range(nrepeat):
        RandLL = pd.read_csv(os.path.join(logpath,'.'.join(['RandLL',name,str(ridx),'csv'])),header=0,index_col=0)
        ll_list.append(RandLL['LL'].values[-1])
    
    ll_argmax = np.argmax(ll_list)
    #print(ll_argmax)

    ridx = ll_argmax
    
    R_est0 = pd.read_csv(os.path.join(logpath,'.'.join(['Rtrack',name,str(ridx),'csv'])),header=0,index_col=0)
    esti_r_list = R_est0.iloc[:,-1].values
    
    #w_corrected = [ratio_correction(1/(r+1),left_mg_pmass[ct][0],left_mg_pmass[ct][1]) for r in esti_r_list]
    #r_list = [(1-w)/w for w in esti_r_list]
    
    return esti_r_list



def estimateW(mg_Alpha_new,Y_new,output,w_init=None,nrepeat=10,name='',n_cores=None):

    ndoublets = len(Y_new)

    Alpha_mat = mg_Alpha_new
    
    alpha0 = Alpha_mat[0]
    alpha1 = Alpha_mat[1]
    
    if w_init is None:
        w_init = initialize_w(Y_new, alpha0, alpha1)
        
    for repeat in range(nrepeat):
    
        log = open(os.path.join(output,'.'.join(['log',name,str(repeat),'txt'])),'w',buffering=1)
    
        t0 = time.time() 
        log.write('start time '+str(t0)+'\n')
    
        # initialize  
        r_init = 2**np.random.normal(np.log2((1-w_init)/w_init),1)
        log.write('initizalized R '+str(r_init)+'\n')
        logR_m_record = []
        LL_record = []
        r_track = []
    
        logR_m = np.log2(r_init)
        logR_std = 1
        logR_m_record.append(logR_m)
    
        num_p = 1000
        num_w = 100
        maxiter = 2
        delta_LL_tol = 0.1
        job_unit = 20
        if n_cores is None:
            n_cores = min(50, int(mp.cpu_count()*0.8))
        
        parallel = True
        if ndoublets <= job_unit:
            parallel = False
    
        if parallel:
            # number of samples for each parallel job in E-step
            num_jobs = ndoublets // job_unit
            jobs_sidx = [i*job_unit for i in range(num_jobs)]
            jobs_sidx.append(ndoublets)
    
        keeptimes = []
        keeptimes_tol = 5
    
        LL = -1e308
        iteration = 0
        LL_record.append(LL)
    
        fp_dll0 = dirichlet.logpdf(alpha0/sum(alpha0), alpha0) 
        fp_dll1 = dirichlet.logpdf(alpha1/sum(alpha1), alpha1) 
    
        while( (len(keeptimes) <  keeptimes_tol) and (iteration < maxiter)):
            # Termination creteria: 
            # ------------------------- (1). iteration times exceed the maximum.
            # ------------------------- (2). the delta_LL is less than the threshold k times in a row.   
    
            # print('iteration',iteration)
            log.write('iteration '+str(iteration)+'\n')
    
            tp0 = time.time()
            if parallel:
    
                pool = mp.Pool(n_cores)  
                result_compact = [pool.apply_async( job_Estep, (Y_new, Alpha_mat, fp_dll0, fp_dll1, logR_m, logR_std,  num_p, num_w, jobs_sidx[i], jobs_sidx[i+1]) ) for i in range(num_jobs)]
                pool.close()
                pool.join()
    
                results = [term.get() for term in result_compact] 
                ll_mat = np.concatenate([term[0] for term in results])  # num_doublet x num_p
                w_mat = np.concatenate([term[1] for term in results])   # num_doublet x num_p
    
            else:
    
                ll_mat, w_mat = job_Estep(Y_new, Alpha_mat, fp_dll0, fp_dll1, logR_m, logR_std,  num_p, num_w, 0, len(Y_new))
    
            log.write('\tE step time:'+str(time.time()-tp0)+'\n')
    
            ### M-step:
            best_w_idx = [np.argmax(term) for term in ll_mat]
            ll_doublet = [ll_mat[idx][val] for idx,val in enumerate(best_w_idx)]
    
            delta_LL = np.sum(ll_doublet) - LL
    
            w_best = [w_mat[idx][val] for idx,val in enumerate(best_w_idx)]
            r_best = [(1-w)/w for w in w_best]
    
            if delta_LL > delta_LL_tol:
    
                LL = np.sum(ll_doublet)
                logR_m = np.mean(np.log2(r_best))
                logR_std = np.std(np.log2(r_best))
    
                logR_m_record.append(logR_m)
                LL_record.append(LL)
                r_track.append(r_best)

                log.write('\testi R '+str(2**logR_m)+'\n\tdelta LL '+str(delta_LL)+'\n')
    
            keeptimes.append(delta_LL < delta_LL_tol)
            if not np.all(keeptimes):
                keeptimes = []    
    
            iteration += 1
    
        tconsuming = time.time() - t0
    
        est_R = 2**logR_m_record[-1]#2**np.median(logR_m_record[2:]) if len(logR_m_record) > 2 else 2**logR_m_record[-1]

        log.write('final R '+str(est_R)+'\n')
        log.write('time '+str(tconsuming)+'\n\n\n')

        log.close()
    
        ## record R_estimation and LL each update
        R_LL_df = pd.DataFrame({'R':2**np.array(logR_m_record),'LL':LL_record})
        R_LL_df.to_csv(os.path.join(output,'.'.join(['RandLL',name,str(repeat),'csv'])))
    
        ## record r estimated for each doublet per update
        r_track_df = pd.DataFrame(np.array(r_track).transpose(),columns=range(len(r_track)))
        r_track_df.to_csv(os.path.join(output,'.'.join(['Rtrack',name,str(repeat),'csv'])))
     


def job_Estep(Y_test, Alpha_mat, fp_dll0, fp_dll1, logR_m, logR_std,  num_p, num_w, start_sidx, end_sidx):
    
    np.random.seed()
    
    sidx_range = range(start_sidx, end_sidx)
    alpha0, alpha1 = Alpha_mat
    
    ybatch_ll_list = []
    ybatch_w_list = []
  
    ### E-setp:
    for i in sidx_range:
        
        y = Y_test[i]
        
        logR_sampling = np.random.normal(logR_m, logR_std, num_w)
        W = 1/(2**logR_sampling+1)
    
        logR_ll = norm.logpdf(logR_sampling, logR_m, logR_std) 
        
        d_p0_sampling  = np.random.dirichlet(alpha0, num_p)
        d_p0_sampling = get_P_valid(d_p0_sampling)
        diri_ll_0 = [dirichlet.logpdf(p, alpha0)  - fp_dll0 for p in d_p0_sampling]
        
        d_p1_sampling  = np.random.dirichlet(alpha1, num_p)
        d_p1_sampling = get_P_valid(d_p1_sampling)
        diri_ll_1 = [dirichlet.logpdf(p, alpha1) - fp_dll1 for p in d_p1_sampling]
                
        y_ll = []
        for w_idx in range(len(W)):
            
            w = W[w_idx]
            p_wsum = d_p0_sampling * w + d_p1_sampling * (1-w)
 
            sum_term = p_wsum.sum(1)
            p_wsum = p_wsum/sum_term.reshape(-1,1)       
            
            mult_ll = (np.log(p_wsum)*y).sum(1)

            pvec_y_ll = np.array(diri_ll_0)+np.array(diri_ll_1)+logR_ll[w_idx]+np.array(mult_ll) 
            #pvel_y_ll_midx = np.argmax(pvec_y_ll)
    
            y_ll.append(np.max(pvec_y_ll))

        ybatch_ll_list.append(y_ll)
        ybatch_w_list.append(W) 
        
  
    ll_mat = np.array(ybatch_ll_list) # ndoublet x num_w
    w_mat = np.array(ybatch_w_list)   # ndoublet x num_w
    
    return ll_mat, w_mat



def get_P_valid(P):
    
    P[P==0] = 1e-10 

    return P
